//lfa/am-fa-gmm.h

#ifndef KALDI_AM_FA_GMM_H_
#define KALDI_AM_FA_GMM_H_ 1

#include <vector>

#include "base/kaldi-common.h"
#include "gmm/model-common.h"
#include "lfa/diag-fagmm.h"
//#include "lfa/fa-gmm.h"
//#include "lfa/full-fa-gmm.h"
#include "itf/options-itf.h"

namespace kaldi {
/// @defgroup DiagGmm DiagGmm
/// @{
/// kaldi Factor Analysis Gaussian Mixture Models

class AmFaGmm {
 public:
  AmFaGmm() {}
  ~AmFaGmm();

  /// Initializes with a single "prototype" GMM.
  void Init(const DiagFaGmm &proto, int32 num_pdfs);
  /// Adds a GMM to the model, and increments the total number of PDFs.
  void AddPdf(const DiagFaGmm &gmm);
  /// Copies the parameters from another model. Allocates necessary memory.
  void CopyFromAmFaGmm(const AmFaGmm &other);

  void SplitPdf(int32 idx, int32 target_components, float perturb_factor);

  // In SplitByCount we use the "target_components" and "power"
  // to work out targets for each state (according to power-of-occupancy rule),
  // and any state less than its target gets mixed up.  If some states
  // were over their target, this may take the #Gauss over the target.
  // we enforce a min-count on Gaussians while splitting (don't split
  // if it would take it below min-count).
  void SplitByCount(const Vector<BaseFloat> &state_occs,
                    int32 target_components, float perturb_factor,
                    BaseFloat power, BaseFloat min_count);


  // In SplitByCount we use the "target_components" and "power"
  // to work out targets for each state (according to power-of-occupancy rule),
  // and any state over its target gets mixed down.  If some states
  // were under their target, this may take the #Gauss below the target.
  void MergeByCount(const Vector<BaseFloat> &state_occs,
                    int32 target_components,
                    BaseFloat power, BaseFloat min_count);

  /// Sets the gconsts for all the PDFs. Returns the total number of Gaussians
  /// over all PDFs that are "invalid" e.g. due to zero weights or variances.
  int32 ComputeGconsts(const FaGmm &fagmm);

  void ComputeExtendMeanBias(const FaGmm &fagmm);
  
  BaseFloat LogLikelihood(const int32 pdf_index,
                          const Matrix<BaseFloat> &data,
			  const std::vector<int32> &gselect,
			  const Vector<BaseFloat> &posteriors) const;
			  
  BaseFloat LogLikelihoodPLDAPoint(const int32 pdf_index, 
				        const FaGmm &fagmm,
					const Vector<double> &data,
				        const Matrix<double> &data_prexform,
					const std::vector<int32> &gselect) const;
					
  void Read(std::istream &in_stream, bool binary);
  void Write(std::ostream &out_stream, bool binary) const;

  int32 Dim() const {
    return (densities_.size() > 0)? densities_[0]->Dim() : 0;
  }
  int32 NumPdfs() const { return densities_.size(); }
  int32 NumGauss() const;
  int32 NumGaussInPdf(int32 pdf_index) const;

  /// Accessors
  DiagFaGmm& GetPdf(int32 pdf_index);
  const DiagFaGmm& GetPdf(int32 pdf_index) const;
  void GetGaussianMean(int32 pdf_index, int32 gauss,
                       VectorBase<BaseFloat> *out) const;
  void GetGaussianVariance(int32 pdf_index, int32 gauss,
                           SpMatrix<BaseFloat> *out) const;

  /// Mutators
  void SetGaussianMean(int32 pdf_index, int32 gauss_index,
                       const VectorBase<BaseFloat> &in);

 private:
  std::vector<DiagFaGmm*> densities_;
 
//  FaGmm fagmm_; //front-end factor analysis GMM
  
  void RemovePdf(int32 pdf_index);

  KALDI_DISALLOW_COPY_AND_ASSIGN(AmFaGmm);
};


inline BaseFloat AmFaGmm::LogLikelihood(const int32 pdf_index, 
					const Matrix<BaseFloat> &data,
					const std::vector<int32> &gselect,
					const Vector<BaseFloat> &posteriors) const {
    
  return densities_[pdf_index]->LogLikelihood(data, gselect, posteriors);
}

inline BaseFloat AmFaGmm::LogLikelihoodPLDAPoint(const int32 pdf_index, 
				        const FaGmm &fagmm,
					const Vector<double> &data,
				        const Matrix<double> &data_prexform,
					const std::vector<int32> &gselect) const {
  Vector<BaseFloat> loglikes;
  return densities_[pdf_index]->LogLikelihoodsPLDAPoint(fagmm, data, data_prexform, gselect, &loglikes);
}

inline void AmFaGmm::ComputeExtendMeanBias(const FaGmm &fagmm) {
  
  for (size_t i = 0; i < densities_.size(); i++)
     densities_[i]->ComputeExtendMeanBias(fagmm);
}

inline int32 AmFaGmm::NumGaussInPdf(int32 pdf_index) const {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
      && (densities_[pdf_index] != NULL));
  return densities_[pdf_index]->NumGauss();
}

inline DiagFaGmm& AmFaGmm::GetPdf(int32 pdf_index) {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
               && (densities_[pdf_index] != NULL));
  return *(densities_[pdf_index]);
}

inline const DiagFaGmm& AmFaGmm::GetPdf(int32 pdf_index) const {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
               && (densities_[pdf_index] != NULL));
  return *(densities_[pdf_index]);
}

/*
inline void AmFaGmm::GetGaussianMean(int32 pdf_index, int32 gauss,
                                       VectorBase<BaseFloat> *out) const {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
      && (densities_[pdf_index] != NULL));
  densities_[pdf_index]->GetComponentMean(gauss, out);
}

inline void AmFaGmm::GetGaussianVariance(int32 pdf_index, int32 gauss,
                                           SpMatrix<BaseFloat> *out) const {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
               && (densities_[pdf_index] != NULL));
  densities_[pdf_index]->GetComponentVar(gauss, out);
}


inline void AmFaGmm::SetGaussianMean(int32 pdf_index, int32 gauss_index,
                                       const VectorBase<BaseFloat> &in) {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
               && (densities_[pdf_index] != NULL));
  densities_[pdf_index]->SetComponentMean(gauss_index, in);
}
*/
/*
inline void AmFaGmm::SplitPdf(int32 pdf_index,
                                           int32 target_components,
                                           float perturb_factor) {
  KALDI_ASSERT((static_cast<size_t>(pdf_index) < densities_.size())
               && (densities_[pdf_index] != NULL));
  densities_[pdf_index]->Split(target_components, perturb_factor);
}
*/


}  // namespace kaldi

/// @} FaGmm
#endif  // KALDI_AM_FA_GMM_H_
