// plda/am-plda.cc


#include <queue>
#include <string>
#include <vector>
using std::vector;

#include "plda/am-plda.h"
#include "gmm/model-common.h"
#include "util/stl-utils.h"
#include "tree/clusterable-classes.h"
#include "tree/cluster-utils.h"

namespace kaldi {

PldaUpdateFlagsType StringToPldaUpdateFlags(std::string str) {
  PldaUpdateFlagsType flags = 0;
  for (const char *c = str.c_str(); *c != '\0'; c++) {
    switch (*c) {
      case 'U': flags |= kPldaChannelProjections; break;
      case 'v': flags |= kPldaVars; break;
      case 'w': flags |= kPldaWeights; break;
      case 'b': flags |= kPldaBias; break;
      case 'G': flags |= kPldaPhoneProjections; break;
      case 'z': flags |= kPldaPhoneVectors; break;
      case 'x': flags |= kPldaChannelPrior; break;
      case 't': flags |= kPldaTransitions; break;
      default: KALDI_ERR << "Invalid element " << CharToString(*c)
                         << " of PldaUpdateFlagsType option string "
                         << str;
    }
  }
  return flags;
}

PldaWriteFlagsType StringToPldaWriteFlags(std::string str) {
  PldaWriteFlagsType flags = 0;
  for (const char *c = str.c_str(); *c != '\0'; c++) {
    switch (*c) {
      case 'g': flags |= kPldaGlobalParams; break;
      case 's': flags |= kPldaStateParams; break;
      case 'n': flags |= kPldaNormalizers; break;
//      case 'u': flags |= kSgmmBackgroundGmms; break;
      case 'a': flags |= kPldaWriteAll; break;
      default: KALDI_ERR << "Invalid element " << CharToString(*c)
                         << " of PldaWriteFlagsType option string "
                         << str;
    }
  }
  return flags;
}


void AmPlda::Read(std::istream &in_stream, bool binary) {
  int32 num_states, feat_dim, num_gauss;
  std::string token;

  ExpectToken(in_stream, binary, "<PLDA>");
  ExpectToken(in_stream, binary, "<NUMSTATES>");
  ReadBasicType(in_stream, binary, &num_states);
  ExpectToken(in_stream, binary, "<DIMENSION>");
  ReadBasicType(in_stream, binary, &feat_dim);
  KALDI_ASSERT(num_states > 0 && feat_dim > 0);

  ReadToken(in_stream, binary, &token);

  while (token != "</PLDA>") {
    if (token == "<SigmaInv>") {
      ExpectToken(in_stream, binary, "<NUMGaussians>");
      ReadBasicType(in_stream, binary, &num_gauss);
      SigmaInv_.resize(num_gauss);
      for (int32 i = 0; i < num_gauss; i++) {
        SigmaInv_[i].Read(in_stream, binary);
      }
    } else if (token == "<G>") {
      ExpectToken(in_stream, binary, "<NUMGaussians>");
      ReadBasicType(in_stream, binary, &num_gauss);
      G_.resize(num_gauss);
      for (int32 i = 0; i < num_gauss; i++) {
        G_[i].Read(in_stream, binary);
      }
    } else if (token == "<U>") {
      ExpectToken(in_stream, binary, "<NUMGaussians>");
      ReadBasicType(in_stream, binary, &num_gauss);
      U_.resize(num_gauss);
      for (int32 i = 0; i < num_gauss; i++) {
        U_[i].Read(in_stream, binary);
      }
    } else if (token == "<CovInv_x>") {
      ExpectToken(in_stream, binary, "<NUMGaussians>");
      ReadBasicType(in_stream, binary, &num_gauss);
      CovInv_x_.resize(num_gauss);
      for (int32 i = 0; i < num_gauss; i++) {
        CovInv_x_[i].Read(in_stream, binary);
      }
    } else if (token == "<b>") {
      b_.Read(in_stream, binary);
    } else if (token == "<mean_x>") {
      mean_x_.Read(in_stream, binary);
    } else if (token == "<LambdaInv>") {
      LambdaInv_.Read(in_stream, binary);
    } else if (token == "<v>") {
      v_.resize(num_states);
      for (int32 j = 0; j < num_states; j++) {
        v_[j].Read(in_stream, binary);
      }
    } else if (token == "<F>") {
      F_.resize(num_states);
      for (int32 j = 0; j < num_states; j++) {
        F_[j].Read(in_stream, binary);
      }
    } else if (token == "<c>") {
      c_.resize(num_states);
      for (int32 j = 0; j < num_states; j++) {
        c_[j].Read(in_stream, binary);
      }
    } else if (token == "<w>") {
      w_.Read(in_stream, binary);
    } else if (token == "<n>") {
      n_.resize(num_states);
      for (int32 j = 0; j < num_states; j++) {
        n_[j].Read(in_stream, binary);
      } 
    } else {
      KALDI_ERR << "Unexpected token '" << token << "' in model file ";
    }
    ReadToken(in_stream, binary, &token);
  }

 // InitChannelPrior();
  ComputeV();
  ComputeVarOffset();
//  ComputeSigmaInv();
  if (n_.empty()) {
    ComputeNormalizers();
  }
 
}

void AmPlda::Write(std::ostream &out_stream, bool binary,
                   PldaWriteFlagsType write_params) const {
  int32 num_states = NumPdfs(),
      feat_dim = FeatureDim(),
      num_gauss = NumGauss();

  WriteToken(out_stream, binary, "<PLDA>");
  if (!binary) out_stream << "\n";
  WriteToken(out_stream, binary, "<NUMSTATES>");
  WriteBasicType(out_stream, binary, num_states);
  WriteToken(out_stream, binary, "<DIMENSION>");
  WriteBasicType(out_stream, binary, feat_dim);
  if (!binary) out_stream << "\n";

  if (write_params & kPldaGlobalParams) {
    WriteToken(out_stream, binary, "<SigmaInv>");
    WriteToken(out_stream, binary, "<NUMGaussians>");
    WriteBasicType(out_stream, binary, num_gauss);
    if (!binary) out_stream << "\n";
    for (int32 i = 0; i < num_gauss; i++) {
      SigmaInv_[i].Write(out_stream, binary);
    }
    WriteToken(out_stream, binary, "<G>");
    WriteToken(out_stream, binary, "<NUMGaussians>");
    WriteBasicType(out_stream, binary, num_gauss);
    if (!binary) out_stream << "\n";
    for (int32 i = 0; i < num_gauss; i++) {
      G_[i].Write(out_stream, binary);
    }
    if (U_.size() != 0) {
      WriteToken(out_stream, binary, "<U>");
      WriteToken(out_stream, binary, "<NUMGaussians>");
      WriteBasicType(out_stream, binary, num_gauss);
      if (!binary) out_stream << "\n";
      for (int32 i = 0; i < num_gauss; i++) {
        U_[i].Write(out_stream, binary);
      }
    }
    if (CovInv_x_.size() != 0) {
      WriteToken(out_stream, binary, "<CovInv_x>");
      WriteToken(out_stream, binary, "<NUMGaussians>");
      WriteBasicType(out_stream, binary, num_gauss);
      if (!binary) out_stream << "\n";
      for (int32 i = 0; i < num_gauss; i++) {
        CovInv_x_[i].Write(out_stream, binary);
      }
    }
    WriteToken(out_stream, binary, "<b>");
    b_.Write(out_stream, binary);
    WriteToken(out_stream, binary, "<mean_x>");
    mean_x_.Write(out_stream, binary);
    WriteToken(out_stream, binary, "<LambdaInv>");
    LambdaInv_.Write(out_stream, binary);
  }

  if (write_params & kPldaStateParams) {
    WriteToken(out_stream, binary, "<v>");
    for (int32 j = 0; j < num_states; j++) {
      v_[j].Write(out_stream, binary);
    }
    
    if (F_.size() != 0) {
      WriteToken(out_stream, binary, "<F>");
      for (int32 j = 0; j < num_states; j++) {
        F_[j].Write(out_stream, binary);
      }
    }
    
    WriteToken(out_stream, binary, "<c>");
    for (int32 j = 0; j < num_states; j++) {
      c_[j].Write(out_stream, binary);
    }
    WriteToken(out_stream, binary, "<w>");
    w_.Write(out_stream, binary);
  }

  if (write_params & kPldaNormalizers) {
    WriteToken(out_stream, binary, "<n>");
    if (n_.empty())
      KALDI_WARN << "Not writing normalizers since they are not present.";
    else
      for (int32 j = 0; j < num_states; j++)
        n_[j].Write(out_stream, binary);
  }

  WriteToken(out_stream, binary, "</PLDA>");
}

void AmPlda::InitializeFromFaGmm(const FaGmm &fa_gmm,
                                   int32 num_states) {
  int32 feature_dim = fa_gmm.FeatureDim();
  int32 chn_dim = fa_gmm.SpeechDim();
  int32 phn_dim = fa_gmm.StateDim();
  int32 num_gauss = fa_gmm.NumGauss();
  
  U_.resize(num_gauss);
  G_.resize(num_gauss);
  b_.Resize(num_gauss, feature_dim);
 // mean_x_.Resize(num_gauss, chn_dim);
  LambdaInv_.Resize(num_gauss, feature_dim);
  w_.Resize(num_states, num_gauss);
  v_.resize(num_states);
  c_.resize(num_states);
 // CovInv_x_.resize(num_gauss);
  for (int32 j = 0; j < num_states; j++) {
    v_[j].Resize(1, phn_dim);
    c_[j].Resize(1);
//    v_[j](0, 0) = 1.0;  // Eq. (26): v_{j1} = [1 0 0 ... 0]
    c_[j](0) = 1.0;     // Eq. (25): c_{j1} = 1.0
    w_.Row(j).CopyFromVec(fa_gmm.weights());
  }
  
  for (int32 i = 0; i < num_gauss; i++) {
    U_[i].Resize(feature_dim, chn_dim);
    U_[i].CopyFromMat(fa_gmm.Projection(i));
    
    G_[i].Resize(feature_dim, phn_dim);
    G_[i].CopyFromMat(fa_gmm.StateProjection(i));
 //   G_[i].Scale(0.1);  // use smaller values of init G
    
//    CovInv_x_[i].Resize(chn_dim);
//    CovInv_x_[i].SetUnit();
    
    b_.Row(i).CopyFromVec(fa_gmm.BiasVector(i));
    LambdaInv_.Row(i).CopyFromVec(fa_gmm.InvCovars(i));
  }
  
  ComputeV();
  ComputeSigmaInv();
  ComputeNormalizers();
}

  ///NOTE: V is stored as its inverse
void AmPlda::ComputeV() {
  
//   Matrix<BaseFloat> I(ChannelDim(), ChannelDim());
   Vector<BaseFloat> inv_var(FeatureDim());
    
 //  I.SetUnit();
   VInv_.resize(NumGauss());
   for(int m=0; m<NumGauss(); m++) {
     inv_var.CopyRowFromMat(LambdaInv_, m);
     VInv_[m].Resize(ChannelDim());
//     VInv_[m].CopyFromSp(CovInv_x_[m]);
     VInv_[m].SetUnit();
     VInv_[m].AddMat2Vec(1.0, U_[m], kTrans, inv_var, 1.0);  //V(m) = I + U(m)^T*\Lambda(m)^(-1)*U
     VInv_[m].InvertDouble();
   }//for
    
}
 
void AmPlda::ComputeVarOffset() {

  SpMatrix<BaseFloat> tmp_mat(FeatureDim());
  var_offset_.Resize(NumGauss(), FeatureDim());
  for(int m=0; m<NumGauss(); m++) {
     tmp_mat.AddMat2Sp(1.0, U_[m], kNoTrans, VInv_[m], 0.0);
     var_offset_.Row(m).CopyDiagFromSp(tmp_mat);  
  }
}

//Make sure that SigmaInv is up-to-date
//may need to call ComputeSigmaInv before this funciton
void AmPlda::ComputeNormalizers() {
  BaseFloat DLog2pi = FeatureDim() * log(2 * M_PI);
  Vector<BaseFloat> log_det_Sigma(NumGauss());

  for (int32 i = 0; i < NumGauss(); i++) {
    try {
      log_det_Sigma(i) = - SigmaInv_[i].LogPosDefDet();
    } catch(...) {
      KALDI_WARN << "Covariance is not positive definite, setting to unit";
      SigmaInv_[i].SetUnit();
      log_det_Sigma(i) = 0.0;
    }
  }
  
  n_.resize(NumPdfs());
  for (int32 j = 0; j < NumPdfs(); j++) {
    n_[j].Resize(NumGauss(), NumSubstates(j));
    Matrix<BaseFloat> mu_jmi(NumSubstates(j), FeatureDim());
    Matrix<BaseFloat> SigmaInv_mu(NumSubstates(j), FeatureDim());
    
    for (int32 i = 0; i < NumGauss(); i++) {    
      // mu_jmi = G_{i} * z_{jm} + b_i
    //  mu_jmi.CopyColsFromVec(b_.Row(i));
      mu_jmi.AddMatMat(1.0, v_[j], kNoTrans, G_[i], kTrans, 0.0);
      SigmaInv_mu.AddMatSp(1.0, mu_jmi, kNoTrans, SigmaInv_[i], 0.0);
    
      for (int32 m = 0; m < NumSubstates(j); m++) {
        // mu_{jmi} * \Sigma_{i}^{-1} * mu_{jmi}
        BaseFloat mu_SigmaInv_mu = VecVec(mu_jmi.Row(m), SigmaInv_mu.Row(m));
        BaseFloat logc = log(c_[j](m));
        
        // eq.(31)
        n_[j](i, m) = logc + log(w_(j, i)) - 0.5 * (log_det_Sigma(i) + DLog2pi
            + mu_SigmaInv_mu);
        {  // Mainly diagnostic code.  Not necessary.
          BaseFloat tmp = n_[j](i, m);
          if (!KALDI_ISFINITE(tmp)) {  // NaN or inf
            KALDI_LOG << "Warning: normalizer for j = " << j << ", m = " << m
                      << ", i = " << i << " is infinite or NaN " << tmp << "= "
                      << (logc) << "+" << log(w_(j, i)) << "+" << (-0.5 *
                          log_det_Sigma(i)) << "+" << (-0.5 * DLog2pi)
                      << "+" << (mu_SigmaInv_mu) << ", setting to finite.";
            n_[j](i, m) = -1.0e+40;  // future work(arnab): get rid of magic number
          }
        }
      }
    }
  }
}

void AmPlda::ComputeSigmaInv() {
    Vector<BaseFloat> inv_var(FeatureDim());
    Matrix<BaseFloat> U_v(FeatureDim(), ChannelDim());
    
    if(SigmaInv_.size() != NumGauss()) 
      SigmaInv_.resize(NumGauss());
    
    for(int m=0; m<NumGauss(); m++) {
      inv_var.CopyRowFromMat(LambdaInv_, m);
      U_v.AddDiagVecMat(1.0, inv_var, U_[m], kNoTrans, 0.0);
      SigmaInv_[m].Resize(FeatureDim());
      SigmaInv_[m].SetZero();
      SigmaInv_[m].AddVec(1.0, inv_var);
      SigmaInv_[m].AddMat2Sp(-1.0, U_v, kNoTrans, VInv_[m], 1.0);
    }
    
}

void AmPlda::ComputePerFrameVars(const VectorBase<BaseFloat> &data,
                           const std::vector<int32> &gselect,
                           BaseFloat logdet_s,
                           PldaPerFrameDerivedVars *per_frame_vars) const {
  KALDI_ASSERT(!n_.empty() && "ComputeNormalizers() must be called.");

//  KALDI_LOG<<FeatureDim();
  if (per_frame_vars->NeedsResizing(gselect.size(),
                                    FeatureDim(),
                                    PhoneDim()))
    per_frame_vars->Resize(gselect.size(), FeatureDim(), PhoneDim());

  per_frame_vars->gselect = gselect;
  per_frame_vars->xt.CopyFromVec(data);

  /*
  for (int32 ki = 0, last = gselect.size(); ki < last; ki++) {
    int32 i = gselect[ki];
    per_frame_vars->xti.Row(ki).CopyFromVec(per_frame_vars->xt);
    if (spk_vars.v_s.Dim() != 0)
      per_frame_vars->xti.Row(ki).AddVec(-1.0, spk_vars.o_s.Row(i));
  }
  */
  
  Vector<BaseFloat> SigmaInv_xt(FeatureDim());
  for (int32 ki = 0, last = gselect.size(); ki < last; ki++) {
    int32 i = gselect[ki];
    per_frame_vars->xti.Row(ki).CopyFromVec(per_frame_vars->xt);
    per_frame_vars->xti.Row(ki).AddVec(-1.0, b_.Row(i));
    Vector<BaseFloat> tmp_vec(per_frame_vars->xti.Row(ki));
//    tmp_vec.AddMatVec(-1.0, U_[i], kNoTrans, mean_x_.Row(i), 1.0);
    SigmaInv_xt.AddSpVec(1.0, SigmaInv_[i], tmp_vec, 0.0);
    // Eq (35): z_{i}(t) = G_{i}^{T} \Sigma_{i}^{-1} x_{i}(t)
    per_frame_vars->zti.Row(ki).AddMatVec(1.0, G_[i], kTrans, SigmaInv_xt, 0.0);
    // Eq.(36): n_{i}(t) = -0.5 x_{i}^{T} \Sigma_{i}^{-1} x_{i}(t)
    per_frame_vars->nti(ki) = -0.5 * VecVec(tmp_vec, SigmaInv_xt) + logdet_s;
  }
}

void AmPlda::PosteriorChannelMean (
			  const PldaPerFrameDerivedVars & per_frame_vars,
			  int32 j,
			  int32 m,
			  int32 i,
			  Vector<BaseFloat> *x_jmit) const {		
    x_jmit->Resize(ChannelDim());
    Vector<BaseFloat> tmp_vec(FeatureDim());
    Vector<BaseFloat> w_jmit(ChannelDim());
    
    tmp_vec.CopyFromVec(per_frame_vars.xt);
    tmp_vec.AddVec(-1.0, b_.Row(i));
    tmp_vec.AddMatVec(-1.0, G_[i], kNoTrans, v_[j].Row(m), 1.0);  //remove the state-based bias
    tmp_vec.MulElements(LambdaInv_.Row(i));
    
    w_jmit.AddMatVec(1.0, U_[i], kTrans, tmp_vec, 0.0); //w_jmit = U_m^T*\Lambda_m^(-1)*(\bar(y_t) - b_m)
//    w_jmit.AddSpVec(1.0, CovInv_x_[i], mean_x_.Row(i), 1.0); //bias from prior distribution of x
    x_jmit->AddSpVec(1.0, VInv_[i], w_jmit, 0.0);  //x_jmit = VInv_i*w_jmit
}
			  
void AmPlda::SplitSubstates(const Vector<BaseFloat> &state_occupancies,
                                   int32 target_nsubstates, BaseFloat perturb,
                                   BaseFloat power, BaseFloat max_cond) {
  // power == p in document.  target_nsubstates == T in document.
  KALDI_ASSERT(state_occupancies.Dim() == NumPdfs());
  int32 tot_n_substates_old = 0;
  int32 phn_dim = PhoneDim();
  std::priority_queue<SubstateCounter> substate_counts;
  vector< SpMatrix<BaseFloat> > H_i;
  SpMatrix<BaseFloat> sqrt_H_sm;
  Vector<BaseFloat> rand_vec(phn_dim), v_shift(phn_dim);

  for (int32 j = 0; j < NumPdfs(); j++) {
    BaseFloat gamma_p = pow(state_occupancies(j), power);
    substate_counts.push(SubstateCounter(j, NumSubstates(j), gamma_p));
    tot_n_substates_old += NumSubstates(j);
  }
  if (target_nsubstates <= tot_n_substates_old || tot_n_substates_old == 0) {
    KALDI_WARN << "Cannot split from " << (tot_n_substates_old) <<
        " to " << (target_nsubstates) << " substates.";
    return;
  }

  ComputeH(&H_i);  // set up that array.
  ComputeSmoothingTermsFromModel(H_i, state_occupancies, &sqrt_H_sm, max_cond);
  H_i.clear();
  sqrt_H_sm.ApplyPow(-0.5);

  for (int32 n_states = tot_n_substates_old;
       n_states < target_nsubstates; n_states++) {
    SubstateCounter state_to_split = substate_counts.top();
    substate_counts.pop();
    state_to_split.num_substates++;
    substate_counts.push(state_to_split);
  }

  while (!substate_counts.empty()) {
    int32 j = substate_counts.top().state_index;
    int32 tgt_n_substates_j = substate_counts.top().num_substates;
    int32 n_substates_j     = NumSubstates(j);
    substate_counts.pop();

    if (n_substates_j == tgt_n_substates_j) continue;

    // Resize v[j] and c[j] to fit new substates
    Matrix<BaseFloat> tmp_v_j(v_[j]);
    v_[j].Resize(tgt_n_substates_j, phn_dim);
    v_[j].Range(0, n_substates_j, 0, phn_dim).CopyFromMat(tmp_v_j);
    tmp_v_j.Resize(0, 0);
    
    Vector<BaseFloat> tmp_c_j(c_[j]);
    c_[j].Resize(tgt_n_substates_j);
    c_[j].Range(0, n_substates_j).CopyFromVec(tmp_c_j);
    tmp_c_j.Resize(0);
/*
    //variance term -Liang
    Matrix<BaseFloat> tmp_F_j(F_[j]);  
    F_[j].Resize(tgt_n_substates_j, phn_dim);
    F_[j].Range(0, n_substates_j, 0, phn_dim).CopyFromMat(tmp_F_j);
    tmp_F_j.Resize(0, 0);
*/   
    // Keep splitting substates until obtaining the desired number
    for (; n_substates_j < tgt_n_substates_j; n_substates_j++) {
      int32 split_substate = std::max_element(c_[j].Data(), c_[j].Data()
          + n_substates_j) - c_[j].Data();

      // c_{jkm} := c_{jmk}' := c_{jkm} / 2
      c_[j](split_substate) = c_[j](n_substates_j) = c_[j](split_substate) / 2;

      // v_{jkm} := +/- split_perturb * H_k^{(sm)}^{-0.5} * rand_vec
      std::generate(rand_vec.Data(), rand_vec.Data() + rand_vec.Dim(),
                    RandGauss);
      v_shift.AddSpVec(perturb, sqrt_H_sm, rand_vec, 0.0);
      v_[j].Row(n_substates_j).CopyFromVec(v_[j].Row(split_substate));
      v_[j].Row(n_substates_j).AddVec(1.0, v_shift);
      v_[j].Row(split_substate).AddVec((-1.0), v_shift);
      
      // Copy the variance
   //   F_[j].Row(n_substates_j).CopyFromVec(F_[j].Row(split_substate));
    }
  }
  KALDI_LOG << "Getting rid of normalizers as they will no longer be valid";

  n_.clear();
  KALDI_LOG << "Split " << (tot_n_substates_old) << " substates to "
      << (target_nsubstates);
}

BaseFloat AmPlda::LogLikelihood(const PldaPerFrameDerivedVars &per_frame_vars,
                                int32 j, BaseFloat log_prune) const {
  KALDI_ASSERT(j < NumPdfs());
  const vector<int32> &gselect = per_frame_vars.gselect;

  // Eq.(37): log p(x(t), m, i|j)  [indexed by j, ki]
  // Although the extra memory allocation of storing this as a
  // matrix might seem unnecessary, we save time in the LogSumExp()
  // via more effective pruning.
  Matrix<BaseFloat> logp_x(gselect.size(), NumSubstates(j));

  for (int32 ki = 0, last = gselect.size();  ki < last; ki++) {
    SubVector<BaseFloat> logp_xi(logp_x, ki);
    int32 i = gselect[ki];
    // for all substates, compute z_{i}^T v_{jm}
    logp_xi.AddMatVec(1.0, v_[j], kNoTrans, per_frame_vars.zti.Row(ki), 0.0);
    logp_xi.AddVec(1.0, n_[j].Row(i));  // for all substates, add n_{jim}
    logp_xi.Add(per_frame_vars.nti(ki));  // for all substates, add n_{i}(t)
  }
  // Eq. (38): log p(x(t)|j) = log \sum_{m, i} p(x(t), m, i|j)
  return logp_x.LogSumExp(log_prune);
}

BaseFloat AmPlda::ComponentPosteriors(const PldaPerFrameDerivedVars &per_frame_vars,
                                int32 j, Matrix<BaseFloat> *post) const{
  KALDI_ASSERT(j < NumPdfs());
  if (post == NULL) KALDI_ERR << "NULL pointer passed as return argument.";
  const vector<int32> &gselect = per_frame_vars.gselect;
  int32 num_gselect = gselect.size();
  post->Resize(num_gselect, NumSubstates(j));

  // Eq.(37): log p(x(t), m, i|j) = z_{i}^T v_{jm} (for all substates)
  post->AddMatMat(1.0, per_frame_vars.zti, kNoTrans, v_[j], kTrans, 0.0);
  for (int32 ki = 0; ki < num_gselect; ki++) {
    int32 i = gselect[ki];
    // Eq. (37): log p(x(t), m, i|j) += n_{jim} + n_{i}(t) (for all substates)
    post->Row(ki).AddVec(1.0, n_[j].Row(i));
    post->Row(ki).Add(per_frame_vars.nti(ki));
  }

  // Eq. (38): log p(x(t)|j) = log \sum_{m, i} p(x(t), m, i|j)
  return post->ApplySoftMax();				  				  
}


///////////////////////////////////////////////////////////////////////////////

template<class Real>
void AmPlda::ComputeH(std::vector< SpMatrix<Real> > *H_i) const {
  KALDI_ASSERT(NumGauss() != 0);
  (*H_i).resize(NumGauss());
  SpMatrix<BaseFloat> H_i_tmp(PhoneDim());
  for (int32 i = 0; i < NumGauss(); i++) {
    (*H_i)[i].Resize(PhoneDim());
   // H_i_tmp.AddMat2Vec(1.0, G_[i], kTrans, LambdaInv_.Row(i), 0.0);
     H_i_tmp.AddMat2Sp(1.0, G_[i], kTrans, SigmaInv_[i], 0.0);
    (*H_i)[i].CopyFromSp(H_i_tmp);
  }
}

// Instantiate the template.
template
void AmPlda::ComputeH(std::vector< SpMatrix<float> > *H_i) const;
template
void AmPlda::ComputeH(std::vector< SpMatrix<double> > *H_i) const;

// Compute the "smoothing" matrices from expected counts given the model.
void AmPlda::ComputeSmoothingTermsFromModel(
    const std::vector< SpMatrix<BaseFloat> > &H,
    const Vector<BaseFloat> &state_occupancies, SpMatrix<BaseFloat> *H_sm,
    BaseFloat max_cond) const {
  int32 num_gauss = NumGauss();
  BaseFloat tot_sum = 0.0;
  KALDI_ASSERT(state_occupancies.Dim() == NumPdfs());
  H_sm->Resize(PhoneDim());
  H_sm->SetZero();
  Vector<BaseFloat> gamma_i(num_gauss);
  gamma_i.SetZero();
  for (int32 j = 0; j < NumPdfs(); j++) {
    int32 M_j = NumSubstates(j);
    KALDI_ASSERT(M_j > 0);
    for (int32 m = 0; m < M_j; m++) {
      gamma_i.AddVec(state_occupancies(j) * c_[j](m), w_.Row(j));
    }
  }
  BaseFloat sum = 0.0;
  for (int32 i = 0; i < num_gauss; i++) {
    if (gamma_i(i) > 0) {
      H_sm->AddSp(gamma_i(i), H[i]);
      sum += gamma_i(i);
    }
  }
  if (sum == 0.0) {
    KALDI_WARN << "Sum of counts is zero. ";
    // set to unit matrix--arbitrary non-singular matrix.. won't ever matter.
    H_sm->SetUnit();
  } else {
    H_sm->Scale(1.0 / sum);
    int32 tmp = H_sm->LimitCondDouble(max_cond);
    if (tmp > 0) {
      KALDI_WARN << "Limited " << (tmp) << " eigenvalues of H_sm";
    }
  }
  tot_sum += sum;

  KALDI_LOG << "ComputeSmoothingTermsFromModel: total count is " << tot_sum;
}

}  // namespace kaldi
