//plda/am-plda.h

#ifndef KALDI_AM_PLDA_H_
#define KALDI_AM_PLDA_H_ 1

#include <vector>

#include "base/kaldi-common.h"
#include "gmm/model-common.h"
#include "lfa/fa-gmm.h"
#include "itf/options-itf.h"

namespace kaldi {

enum PldaUpdateFlags {
    kPldaChannelProjections     	= 0x001,  // U
    kPldaVars          			= 0x002,  // lambda
    kPldaWeights       		= 0x004,  // w
    kPldaBias          			= 0x008,  // b
    kPldaPhoneProjections 	= 0x010,  // G
    kPldaPhoneVectors		= 0x020,  // z
    kPldaChannelPrior		= 0x040,
    kPldaTransitions			= 0x080  // t 
  };

typedef uint16 PldaUpdateFlagsType;

PldaUpdateFlagsType StringToPldaUpdateFlags(std::string str);

enum PldaWriteFlags {
  kPldaGlobalParams    	= 0x001,  /// g
  kPldaStateParams     	= 0x002,  /// s
  kPldaNormalizers      	= 0x004,  /// n
  kPldaWriteAll        	    	= 0x00F  /// a
};

typedef uint16 PldaWriteFlagsType;  ///< Bitwise OR of the above flags.

PldaWriteFlagsType StringToPldaWriteFlags(std::string str);


/** \struct PldaPerFrameDerivedVars
 *  Holds the per-frame precomputed quantities x(t), x_{i}(t), z_{i}(t), and
 *  n_{i}(t) for the AmPlda, as well as the cached Gaussian
 *  selection records.
 */
struct PldaPerFrameDerivedVars {
  std::vector<int32> gselect;
  Vector<BaseFloat> xt;   ///< x'(t), FMLLR-adapted, dim = [D], eq.(33)
  Matrix<BaseFloat> xti;  ///< x_{i}(t) = x'(t) - b_i: dim = [I][D], eq.(34)
  Matrix<BaseFloat> zti;  ///< z_{i}(t), dim = [I][S], eq.(35)
  Vector<BaseFloat> nti;  ///< n_{i}(t), dim = [I], eq.(36)

  PldaPerFrameDerivedVars() : xt(0), zti(0, 0), nti(0) {}
  void Resize(int32 ngauss, int32 feat_dim, int32 phn_dim) {
    xt.Resize(feat_dim);
    xti.Resize(ngauss, feat_dim);
    zti.Resize(ngauss, phn_dim);
    nti.Resize(ngauss);
  }
  bool IsEmpty() const {
    return (xt.Dim() == 0 || xti.NumRows() == 0 || zti.NumRows() == 0 || nti.Dim() == 0);
  }
  bool NeedsResizing(int32 ngauss, int32 feat_dim, int32 phn_dim) const {
    return (xt.Dim() != feat_dim || zti.NumRows() != ngauss || xti.NumRows() != ngauss
        || zti.NumCols() != phn_dim || xti.NumCols() != feat_dim || nti.Dim() != ngauss);
  }
};

/**\SubstateCounter
*  Used for sub-state splitting
*/
struct SubstateCounter {
  SubstateCounter(int32 j, int32 num_substates, BaseFloat occ)
      : state_index(j), num_substates(num_substates), occupancy(occ) {}

  int32 state_index;
  int32 num_substates;
  BaseFloat occupancy;

  bool operator < (const SubstateCounter &r) const {
    return occupancy/num_substates < r.occupancy/r.num_substates;
  }
};


/** \class AmPlda
 *  Class for definition of the plda acoustic model
 */
class AmPlda {
 public:
  AmPlda() {}
  //~AmPlda();

    
  void Read(std::istream &in_stream, bool binary);
  void Write(std::ostream &out_stream, bool binary, 
	     PldaWriteFlagsType write_params) const;
  
  void InitializeFromFaGmm(const FaGmm &fa_gmm, int32 num_states);
  
  /// Copies the parameters from another model. Allocates necessary memory.
  void CopyFromAmPlda(const AmPlda &other);

  void SplitPdf(int32 idx, int32 target_components, float perturb_factor);

  // In SplitByCount we use the "target_components" and "power"
  // to work out targets for each state (according to power-of-occupancy rule),
  // and any state less than its target gets mixed up.  If some states
  // were over their target, this may take the #Gauss over the target.
  // we enforce a min-count on Gaussians while splitting (don't split
  // if it would take it below min-count).
  void SplitByCount(const Vector<BaseFloat> &state_occs,
                    int32 target_components, float perturb_factor,
                    BaseFloat power, BaseFloat min_count);

  /// This needs to be called with each new frame of data, prior to accumulation
  /// or likelihood evaluation: it computes various pre-computed quantities. The
  /// 'logdet_s' term is the log determinant of the FMLLR transform, or 0.0 if
  /// no FMLLR is used or it's single-class fMLLR applied in the feature
  /// extraction, and we're not keeping track of it here.
  void ComputePerFrameVars(const VectorBase<BaseFloat> &data,
                           const std::vector<int32> &gselect,
                           BaseFloat logdet_s,							
                           PldaPerFrameDerivedVars *per_frame_vars) const;
			   
  void  PosteriorChannelMean(
			  const PldaPerFrameDerivedVars & per_frame_vars,
			  int32 j,
			  int32 m,
			  int32 i,
			  Vector<BaseFloat> *x_jmit) const;		   
							  
  /// Computes the data-independent terms in the log-likelihood computation
  /// for each Gaussian component and all substates.
  void ComputeNormalizers();	
  
  /// Increases the total number of substates based on the state occupancies.
  void SplitSubstates(const Vector<BaseFloat> &state_occupancies,
                      int32 target_nsubstates,
                      BaseFloat perturb,
                      BaseFloat power, 
                      BaseFloat cond);
		      
  /// This does a likelihood computation for a given state using the
  /// top-scoring Gaussian components (in per_frame_vars).  If the
  /// log_prune parameter is nonzero (e.g. 5.0), the LogSumExp() stage is
  /// pruned, which is a significant speedup... smaller values are faster.
  BaseFloat LogLikelihood(const PldaPerFrameDerivedVars &per_frame_vars,
                          int32 state_index, BaseFloat log_prune = 0.0) const;
			  
  /// Similar to LogLikelihood() function above, but also computes the posterior
  /// probabilities for the top-scoring Gaussian components and all substates.
  BaseFloat ComponentPosteriors(const PldaPerFrameDerivedVars &per_frame_vars,
                                int32 state, Matrix<BaseFloat> *post) const;				
  
  /// Compute the inverse covariance matrices for likelihood evaluation 
  /// SigmaInv_ = (UU^T + Lambda_)^{-1}
  void ComputeSigmaInv();
  
  /// Compuate VInv_ : (I+U_m^T\LambdaInv_m U_m)^{-1}
  void ComputeV();
  void ComputeVarOffset();
  void InitChannelPrior() {
      mean_x_.Resize(NumGauss(), ChannelDim());
      mean_x_.SetZero();
      CovInv_x_.resize(NumGauss());
      for(int32 i=0; i<NumGauss(); i++) {
        CovInv_x_[i].Resize(ChannelDim());
	CovInv_x_[i].SetUnit();
      }
  }
  
  int32 NumPdfs() const {  return c_.size(); }
  int32 NumGauss() const { return U_.size(); }  
  int32 NumSubstates(int32 j) const { return c_[j].Dim(); }
  int32 PhoneDim() const { return G_[0].NumCols(); }
  int32 ChannelDim() const { return U_[0].NumCols(); }
  int32 FeatureDim() const { return U_[0].NumRows(); }
  int32 NumActiveWeights(BaseFloat threshold) const { 
    int32 cnt = 0;
    for(int32 j=0; j<NumPdfs(); j++) 
      for(int32 m=0; m<NumGauss(); m++)
	if(w_(j,m) > threshold) cnt++;
	
	return cnt;
  }
  
  /// Accessors
   const Matrix<BaseFloat>& BiasVectors() const { return b_; }
      
   const Matrix<BaseFloat>& ChannelProjection(int32 m) const {
      KALDI_ASSERT(m<NumGauss());
      return U_[m];
   }

   const Matrix<BaseFloat>& PhoneProjection(int32 m) const {
      KALDI_ASSERT(m<NumGauss());
      return G_[m];
    }
  
  const Matrix<BaseFloat>& StateVectors(int32 j) const {
    return v_[j];
  }
  
   const SpMatrix<BaseFloat>& InverseV(int32 m) const {
      KALDI_ASSERT(m<NumGauss());
      return VInv_[m];
   }
   
   const SpMatrix<BaseFloat>& SigmaInv(int32 m) const {
      KALDI_ASSERT(m<NumGauss());
      return SigmaInv_[m];
   }
   
   const Matrix<BaseFloat>& LambdaInv() const {
      return LambdaInv_;
   }
   
   const Matrix<BaseFloat> & VarOffset() const {
     return var_offset_;  
   }
   
  /// Computes quantities H = G_i Lambda_i^{-1} G_i^T.
  template<class Real>
  void ComputeH(std::vector< SpMatrix<Real> > *H_i) const;

  private:
    
    void ComputeSmoothingTermsFromModel(
    const std::vector< SpMatrix<BaseFloat> > &H,
    const Vector<BaseFloat> &state_occupancies, SpMatrix<BaseFloat> *H_sm,
    BaseFloat max_cond) const;
  
 private:
  
   ///covariance projection matrix, dim [I][D][S]
   std::vector< Matrix<BaseFloat> > U_;
   ///state-related projection matrix, dim[I][D][T]
   std::vector< Matrix<BaseFloat> > G_;
   ///offset mean vectors, stacked in a matrix, dim [I][D]
   Matrix<BaseFloat> b_;
   ///residual inverse diagonal covariance matrix, stacked in a matrix, dim [I][D]
   Matrix<BaseFloat> LambdaInv_;
   ///Sigma_Inv_ = Invert(I + UUT), will apply the matrix inversion lemma to efficiency
   std::vector< SpMatrix<BaseFloat> > SigmaInv_;   
   ///posterior per-component variance matrix, 
   ///NOTE there are not precision matrices, dim [I][S][S]
   std::vector< SpMatrix<BaseFloat> > VInv_; 
   ///diag(UVInvU^T), used for covariance update
   Matrix<BaseFloat> var_offset_;
   ///mean vectors of observation variable x_{jkmt}
   Matrix<BaseFloat> mean_x_;
   ///inverse covariance matrices of observation variable x_{jkmt}
   std::vector< SpMatrix<BaseFloat> > CovInv_x_;
   
   Matrix<BaseFloat> w_;        ///< weights (not log). Dimension is [J][M]
   std::vector< Matrix<BaseFloat> > v_;       ///< sub-state vectors
   std::vector< Matrix<BaseFloat> > F_;    ///< diagonal variance of sub-state vectors
   
   /// c_{k}, mixture weights. Dimension is [J][M_{j}]
  std::vector< Vector<BaseFloat> > c_;
   /// n_{jmk}, per-Gaussian normalizer. Dimension is [J][M][K_j}]
   std::vector< Matrix<BaseFloat> > n_;
  

  KALDI_DISALLOW_COPY_AND_ASSIGN(AmPlda);
  friend class MleAmPldaUpdater;
  friend class EbwAmPldaUpdater;
};


}  // namespace kaldi

#endif  // KALDI_AM_PLDA_H_
