// plda/ebw-am-plda.cc

// Liang Lu (2014)

// See ../../COPYING for clarification regarding multiple authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.

#include "base/kaldi-common.h"
#include "plda/ebw-am-plda.h"
//#include "thread/kaldi-thread.h"
using std::vector;

namespace kaldi {

void EbwAmPldaUpdater::Update(const MleAmPldaAccs &num_accs,
                              const MleAmPldaAccs &den_accs,
                              AmPlda *model,
                              PldaUpdateFlagsType flags,
                              BaseFloat *auxf_change_out,
                              BaseFloat *count_out) {

  KALDI_ASSERT((flags & (kPldaPhoneVectors | kPldaPhoneProjections |
                         kPldaChannelProjections | kPldaVars | kPldaBias | kPldaWeights )) != 0);

  // Various quantities need to be computed at the start, before we
  // change any of the model parameters.
  std::vector< SpMatrix<double> > Q_num, Q_den, H;
  
  if (flags & kPldaPhoneProjections) {
    MleAmPldaUpdater::ComputeQ(num_accs, *model, &Q_num);
    MleAmPldaUpdater::ComputeQ(den_accs, *model, &Q_den);
  }
  
  if (flags & kPldaPhoneVectors)
    model->ComputeH(&H);

  BaseFloat tot_impr = 0.0;

  if (flags & kPldaPhoneVectors)
    tot_impr += UpdatePhoneVectors(num_accs, den_accs, model, H);
  
  if (flags & kPldaPhoneProjections)
    tot_impr += UpdateG(num_accs, den_accs, Q_num, Q_den, model);

  if (flags & kPldaChannelProjections)
    tot_impr += UpdateU(num_accs, den_accs, model);
  
  if (flags & kPldaVars)
    tot_impr += UpdateVars(num_accs, den_accs, model);

  if (flags & kPldaBias)
    tot_impr += UpdateBias(num_accs, den_accs, model);
  
  if (flags & kPldaWeights) {
    tot_impr += UpdateSubstateWeights(num_accs, den_accs, model);
    tot_impr += UpdateComponentWeights(num_accs, den_accs, model);
  }

  if (auxf_change_out) *auxf_change_out = tot_impr * num_accs.total_frames_;
  if (count_out) *count_out = num_accs.total_frames_;
  
  if (fabs(num_accs.total_frames_ - den_accs.total_frames_) >
      0.01*(num_accs.total_frames_ + den_accs.total_frames_))
    KALDI_WARN << "Num and den frame counts differ, "
               << num_accs.total_frames_ << " vs. " << den_accs.total_frames_;

  BaseFloat like_diff = num_accs.total_like_ - den_accs.total_like_;
  
  KALDI_LOG << "***Averaged differenced likelihood per frame is "
            << (like_diff/num_accs.total_frames_)
            << " over " << (num_accs.total_frames_) << " frames.";
  KALDI_LOG << "***Note: for this to be at all meaningful, if you use "
            << "\"canceled\" stats you will have to renormalize this over "
            << "the \"real\" frame count.";

   if(flags & (kPldaChannelProjections | kPldaVars |kPldaChannelPrior)) {
     model->ComputeV();  //update the matrix inv_V
   }
   
   model->ComputeSigmaInv();
   model->ComputeNormalizers();
}

double EbwAmPldaUpdater::UpdatePhoneVectors(
    const MleAmPldaAccs &num_accs,
    const MleAmPldaAccs &den_accs,    
    AmPlda *model,
    const std::vector<SpMatrix<double> > &H) const {

  int32 S = num_accs.phn_dim_;
  double tot_impr = 0.0, count = 0.0;
  
  for (int32 j = 0; j < model->NumPdfs(); j++) {
    double num_state_count = 0.0,
        state_auxf_impr = 0.0;
    for (int32 m = 0; m < model->NumSubstates(j); m++) {
      double gamma_jm_num = num_accs.gamma_[j].Row(m).Sum();
      double gamma_jm_den = den_accs.gamma_[j].Row(m).Sum();
      num_state_count += gamma_jm_num;
      Vector<double> g_jm_num(S);  // computed using eq. 58 of SGMM paper [for numerator stats]
      SpMatrix<double> H_jm_num(S);  // computed using eq. 59 of SGMM paper [for numerator stats]
      Vector<double> g_jm_den(S); // same, but for denominator stats.
      SpMatrix<double> H_jm_den(S);
      
      g_jm_num.CopyFromVec(num_accs.y_[j].Row(m));  
      for (int32 i = 0; i < num_accs.num_gauss_; i++) {
        double gamma_jmi = num_accs.gamma_[j](m, i);
        if (gamma_jmi != 0.0)
          H_jm_num.AddSp(gamma_jmi, H[i]);
      }
     
      //KALDI_LOG<<gamma_jm_num<<" "<<gamma_jm_den<<" "<<num_state_count;
  
      //smoothing
      //g_jm_num.Scale((gamma_jm_num+options_.tau_v)/gamma_jm_num);
      //H_jm_num.Scale((gamma_jm_num+options_.tau_v)/gamma_jm_num);
      
      g_jm_den.CopyFromVec(den_accs.y_[j].Row(m));  
      for (int32 i = 0; i < den_accs.num_gauss_; i++) {
        double gamma_jmi = den_accs.gamma_[j](m, i);
        if (gamma_jmi != 0.0)
          H_jm_den.AddSp(gamma_jmi, H[i]);
      }
      
      Vector<double> v_jm(model->v_[j].Row(m));
      Vector<double> local_derivative(S); // difference of derivative of numerator
      // and denominator objetive function.
      local_derivative.AddVec(1.0, g_jm_num);
      //local_derivative.AddVec(2.0*gamma_jm_den, v_jm);  //D-term
      local_derivative.AddSpVec(-1.0, H_jm_num, v_jm, 1.0);
      local_derivative.AddVec(-1.0, g_jm_den);
      local_derivative.AddSpVec(-1.0 * -1.0, H_jm_den, v_jm, 1.0);
      
      SpMatrix<double> quadratic_term(H_jm_num);
      quadratic_term.AddSp(1.0, H_jm_den);
      double substate_count = 1.0e-10 + gamma_jm_num + gamma_jm_den;
      quadratic_term.Scale( (substate_count + options_.tau_v) / substate_count);
      quadratic_term.Scale(1.0 / (options_.lrate_v + 1.0e-10) );

   //   local_derivative.Scale(0.1);
   //   quadratic_term.Scale(0.1);

  //      if(j == 5421) {
  //	KALDI_LOG<<gamma_jm_num<<" "<<gamma_jm_den<<" "<<num_state_count;
     //   KALDI_LOG<<local_derivative<<quadratic_term;
   //    }

      Vector<double> delta_v_jm(S);

      SolverOptions opts;
      opts.name = "v";
      opts.K = options_.max_cond;
      opts.eps = options_.epsilon;
      opts.diagonal_precondition = true;
      
      double auxf_impr =
          ((gamma_jm_num + gamma_jm_den == 0) ? 0.0 :
           SolveQuadraticProblem(quadratic_term, local_derivative,
                                 opts, &delta_v_jm));

      v_jm.AddVec(1.0, delta_v_jm);
      model->v_[j].Row(m).CopyFromVec(v_jm);
      state_auxf_impr += auxf_impr;
    }

    tot_impr += state_auxf_impr;
    count += num_state_count;
    
    if (j < 10) {
      KALDI_LOG << "Objf impr for state j = " << j << "  is "
                << (state_auxf_impr / (num_state_count + 1.0e-10))
                << " over " << num_state_count << " frames";
    }
  }
  
   tot_impr /= count;

  KALDI_LOG << "**Overall auxf improvement for v is " << tot_impr
            << " over " << count << " frames";
  return tot_impr;
}

double EbwAmPldaUpdater::UpdateG(const MleAmPldaAccs &num_accs,
                                 const MleAmPldaAccs &den_accs,
                                 const std::vector< SpMatrix<double> > &Q_num,
                                 const std::vector< SpMatrix<double> > &Q_den,
                                 AmPlda *model) const {
  int32 S = model->PhoneDim(),
      D = model->FeatureDim(),
      I = model->NumGauss();
  
  Vector<double> num_count_vec(I), den_count_vec(I), impr_vec(I);
  for (int32 j = 0; j < num_accs.num_states_; j++) {
    num_count_vec.AddRowSumMat(1.0, num_accs.gamma_[j]);
    den_count_vec.AddRowSumMat(1.0, den_accs.gamma_[j]);
  }

  for (int32 i = 0; i < I; i++) {
    double gamma_i_num = num_count_vec(i), gamma_i_den = den_count_vec(i);

    if (gamma_i_num + gamma_i_den == 0.0) {
      KALDI_WARN << "Not updating phonetic basis for i = " << i
                 << " because count is zero. ";
      continue;
    }
    
    double smoothing_rate = (gamma_i_num + options_.tau_G)/gamma_i_num;
    Matrix<double> Gi(model->G_[i]);    
    Matrix<double> L(D, S); // this is something like the Y quantity, which
    // represents the linear term in the objf on M-- except that we make it the local
    // derivative about the current value, instead of the derivative around zero.
    // But it's not exactly the derivative w.r.t. M, due to the factor of Sigma_i.
    // The auxiliary function is Q(x) = tr(M^T P Y) - 0.5 tr(P M Q M^T),
    // where P is Y^{-1}.  The quantity L we define here will be Y - M Q,
    // and you can think of this as like the local derivative, except there is
    // a term P in there.
    L.AddMat(smoothing_rate, num_accs.G_accumulator_[i]);
    L.AddMatSp(-1.0*smoothing_rate, Gi, kNoTrans, Q_num[i], 1.0);
    L.AddMat(-1.0, den_accs.G_accumulator_[i]);
    L.AddMatSp(-1.0*-1.0, Gi, kNoTrans, Q_den[i], 1.0);

    SpMatrix<double> Q(S); // This is a combination of the Q's for the numerator and denominator.
    Q.AddSp(smoothing_rate, Q_num[i]);
    Q.AddSp(1.0, Q_den[i]);

    double state_count = 1.0e-10 + options_.tau_G + gamma_i_num + gamma_i_den; // the count
    // represented by the quadratic part of the stats.
//    Q.Scale( (state_count + options_.tau_G) / state_count );
    Q.Scale( 1.0 / (options_.lrate_G + 1.0e-10) );

    SolverOptions opts;
    opts.name = "G";
    opts.K = options_.max_cond;
    opts.eps = options_.epsilon;
    
    SpMatrix<double> this_var(model->FeatureDim());
    this_var.AddVec(1.0, model->LambdaInv().Row(i));
    Matrix<double> deltaG(D, S);
    double impr = SolveQuadraticMatrixProblem(Q, L, this_var, opts, &deltaG);

    impr_vec(i) = impr;
    Gi.AddMat(1.0, deltaG);
    model->G_[i].CopyFromMat(Gi);
    if (i < 10 || impr / state_count > 3.0) {
      KALDI_LOG << "Objf impr for projection G for i = " << i << ", is "
                << (impr/(gamma_i_num + 1.0e-20)) << " over " << gamma_i_num
                << " frames";
    }
  }
  BaseFloat tot_count = num_count_vec.Sum(), tot_impr = impr_vec.Sum();
  
  tot_impr /= (tot_count + 1.0e-20);
  KALDI_LOG << "Overall auxiliary function improvement for model projections "
            << "G is " << tot_impr << " over " << tot_count << " frames";

  KALDI_VLOG(1) << "Updating G: num-count is " << num_count_vec;
  KALDI_VLOG(1) << "Updating G: den-count is " << den_count_vec;
  KALDI_VLOG(1) << "Updating G: objf-impr is " << impr_vec;
  
  return tot_impr;
}


double EbwAmPldaUpdater::UpdateU(const MleAmPldaAccs &num_accs,
                                 const MleAmPldaAccs &den_accs,
                                 AmPlda *model) const {
  if (num_accs.chn_dim_ == 0 || num_accs.U_accumulator_first_.size() == 0 || 
      num_accs.U_accumulator_second_.size() == 0) {
    KALDI_ERR << "Channel subspace dim is zero or no stats accumulated";
  }

  int32 I = num_accs.num_gauss_, D = num_accs.feature_dim_,
      T = num_accs.chn_dim_;
  
  Vector<double> num_count_vec(I), den_count_vec(I), impr_vec(I);
  for (int32 j = 0; j < num_accs.num_states_; j++) {
    num_count_vec.AddRowSumMat(1.0, num_accs.gamma_[j]);
    den_count_vec.AddRowSumMat(1.0, den_accs.gamma_[j]);
  }
  
  for (int32 i = 0; i < I; i++) {
    double gamma_i_num = num_count_vec(i), gamma_i_den = den_count_vec(i);
    if (gamma_i_num + gamma_i_den == 0.0) {
      KALDI_WARN << "Not updating channel projections for i = " << i
                 << " because count is zero. ";
      continue;
    }
    Matrix<double> Ui(model->U_[i]);
    // See comment near declaration of L in UpdateM().  This update is the
    // same, but change M->N, Y->Z and Q->R.

    double smoothing_rate = (gamma_i_num + options_.tau_U)/gamma_i_num;
    Matrix<double> L(D, T);
    L.AddMat(smoothing_rate, num_accs.U_accumulator_first_[i]);
    L.AddMatSp(-1.0*smoothing_rate, Ui, kNoTrans, num_accs.U_accumulator_second_[i], 1.0);
    L.AddMat(-1.0, den_accs.U_accumulator_first_[i]);
    L.AddMatSp(-1.0*-1.0, Ui, kNoTrans, den_accs.U_accumulator_second_[i], 1.0);
    
    SpMatrix<double> R(T); // combination of the numerator and denominator R's.
    R.AddSp(smoothing_rate, num_accs.U_accumulator_second_[i]);
    R.AddSp(1.0, den_accs.U_accumulator_second_[i]);

    double state_count = 1.0e-10 + options_.tau_U + gamma_i_num + gamma_i_den; // the count
    // represented by the quadratic part of the stats.
 //   R.Scale( (state_count + options_.tau_U) / state_count );
    R.Scale( 1.0 / (options_.lrate_U + 1.0e-10) );
    
    Matrix<double> deltaU(D, T);

    SolverOptions opts;
    opts.name = "U";
    opts.K = options_.max_cond;
    opts.eps = options_.epsilon;

    SpMatrix<double> this_var(model->FeatureDim());
    this_var.AddVec(1.0, model->LambdaInv().Row(i));
    double impr = SolveQuadraticMatrixProblem(R, L, this_var, opts, &deltaU);

    impr_vec(i) = impr;
    Ui.AddMat(1.0, deltaU);
    model->U_[i].CopyFromMat(Ui);
    if (i < 10 || impr / (state_count+1.0e-20) > 3.0) {
      KALDI_LOG << "Objf impr for channel projection U for i = " << (i)
                << ", is " << (impr / (gamma_i_num + 1.0e-20)) << " over "
                << gamma_i_num << " frames";
    }
  }

  KALDI_VLOG(1) << "Updating U: numerator count is " << num_count_vec;
  KALDI_VLOG(1) << "Updating U: denominator count is " << den_count_vec;
  KALDI_VLOG(1) << "Updating U: objf-impr is " << impr_vec;
  
  double tot_count = num_count_vec.Sum(), tot_impr = impr_vec.Sum();
  tot_impr /= (tot_count + 1.0e-20);
  KALDI_LOG << "**Overall auxf impr for U is " << tot_impr
            << " over " << tot_count << " frames";
  return tot_impr;
}

double EbwAmPldaUpdater::UpdateVars(const MleAmPldaAccs &num_accs,
                                    const MleAmPldaAccs &den_accs,
                                    AmPlda *model) const {
  int32 I = num_accs.num_gauss_;
  
  Vector<double> num_count_vec(I), den_count_vec(I), impr_vec(I);
  for (int32 j = 0; j < num_accs.num_states_; j++) {
    num_count_vec.AddRowSumMat(1.0, num_accs.gamma_[j]);
    den_count_vec.AddRowSumMat(1.0, den_accs.gamma_[j]);
  }

  for (int32 i = 0; i < I; i++) {
    double num_count = num_count_vec(i), den_count = den_count_vec(i);

    Vector<double> VarStats(num_accs.var_accumulator_.Row(i));
    VarStats.AddVec(-1.0, den_accs.var_accumulator_.Row(i));
    // SigmaStats now contain the stats for estimating Sigma (as in the main SGMM paper),
    // differenced between num and den.
    Vector<double> LambdaInvOld(model->LambdaInv_.Row(i)), LambdaOld(model->LambdaInv_.Row(i));
    LambdaOld.InvertElements();
    double count = num_count - den_count;
    KALDI_ASSERT(options_.lrate_var <= 1.0);
    double inv_lrate = 1.0 / options_.lrate_var;
    // These formulas assure that the objective function behaves in
    // a roughly symmetric way w.r.t. num and den counts.
    double E_den = 1.0 + inv_lrate, E_num = inv_lrate - 1.0;

    double smoothing_count =
        (options_.tau_var * inv_lrate) + // multiply tau_var by inverse-lrate
        (E_den * den_count) +              // for compatibility with other updates.
        (E_num * num_count) +
        1.0e-10;
	
//     if(i<10)
//      KALDI_LOG<<"smoothing count is " <<smoothing_count<<" count is "<<count;
    
    VarStats.AddVec(smoothing_count, LambdaOld);
    count += smoothing_count;
    VarStats.Scale(1.0 / count);
    Vector<double> LambdaInv(VarStats); // before floor and ceiling.  Currently sigma,
    // not its inverse.
    Vector<double> covfloor(LambdaOld);
    covfloor.Scale(options_.cov_min_value);
    int n_floor = LambdaInv.ApplyFloor(covfloor);
    LambdaInv.InvertElements(); // make it inverse variance.
    covfloor.CopyFromVec(LambdaInvOld);
    covfloor.Scale(options_.cov_min_value);
    int n_ceiling = LambdaInv.ApplyFloor(covfloor);
    
    // this auxf_change.  
    Vector<double> LambdaInv_log(LambdaInv);
    Vector<double> LambdaInvOld_log(LambdaInvOld);
    LambdaInv_log.ApplyLog();
    LambdaInvOld_log.ApplyLog();
    double auxf_change = -0.5 * count *(VecVec(LambdaInv, VarStats)
                                        - VecVec(LambdaInvOld, VarStats)
                                        - LambdaInv_log.Sum()
                                        + LambdaInvOld_log.Sum());

    model->LambdaInv_.Row(i).CopyFromVec(LambdaInv);
    impr_vec(i) = auxf_change;
    if (i < 10 || auxf_change / (num_count+den_count+1.0e-10) > 2.0
        || n_floor+n_ceiling > 0) {
      KALDI_LOG << "Updating variance: Auxf change per frame for Gaussian "
                << i << " is " << (auxf_change / num_count) << " over "
                << num_count << " frames " << "(den count was " << den_count
                << "), #floor,ceil was " << n_floor << ", " << n_ceiling;
    }
  }
  KALDI_VLOG(1) << "Updating Lambda: numerator count is " << num_count_vec;
  KALDI_VLOG(1) << "Updating Lambda: denominator count is " << den_count_vec;
  KALDI_VLOG(1) << "Updating Lambda: objf-impr is " << impr_vec;
  
  double tot_count = num_count_vec.Sum(), tot_impr = impr_vec.Sum();
  tot_impr /= tot_count+1.0e-20;
  KALDI_LOG << "**Overall auxf impr for Lambda is " << tot_impr
            << " over " << tot_count << " frames";
  return tot_impr;
}

double EbwAmPldaUpdater::UpdateBias(const MleAmPldaAccs &num_accs,
                                    const MleAmPldaAccs &den_accs,
                                    AmPlda *model) const {
  int32 I = num_accs.num_gauss_;
  
  Vector<double> num_count_vec(I), den_count_vec(I), impr_vec(I);
  for (int32 j = 0; j < num_accs.num_states_; j++) {
    num_count_vec.AddRowSumMat(1.0, num_accs.gamma_[j]);
    den_count_vec.AddRowSumMat(1.0, den_accs.gamma_[j]);
  }

  for (int32 i = 0; i < I; i++) {
    double num_count = num_count_vec(i), den_count = den_count_vec(i);

    Vector<double> BiasStats(num_accs.b_accumulator_.Row(i));
    BiasStats.AddVec(-1.0, den_accs.b_accumulator_.Row(i));
    
    Vector<double> bi_Old(model->b_.Row(i));
    double count = num_count - den_count;
    KALDI_ASSERT(options_.lrate_b <= 1.0);
    double bias_lrate = 1.0 / options_.lrate_b;
    // These formulas assure that the objective function behaves in
    // a roughly symmetric way w.r.t. num and den counts.
    double E_den = 1.0 + bias_lrate, E_num = bias_lrate - 1.0;

    double smoothing_count =
        (options_.tau_b * bias_lrate) + // multiply tau_var by inverse-lrate
        (E_den * den_count) +              // for compatibility with other updates.
        (E_num * num_count) +
        1.0e-10;

 //   if(i<10)
 //     KALDI_LOG<<"smoothing count is " <<smoothing_count<<" count is "<<count;
    
    BiasStats.AddVec(smoothing_count, bi_Old);
    count += smoothing_count;
  //  BiasStats.Scale(1.0 / count);
    Vector<double> bi(BiasStats); 
    bi.Scale(1.0/count);
    
    // this auxf_change. 
    Vector<double> b_lambda_inv(model->FeatureDim());
    Vector<double> LambdaInv_i(model->LambdaInv_.Row(i));
    b_lambda_inv.AddVecVec(1.0, bi_Old, LambdaInv_i, 0.0);
    double auxf_old = -0.5 * count * VecVec(bi_Old, b_lambda_inv)
  			      + VecVec(b_lambda_inv, BiasStats);
   
    b_lambda_inv.AddVecVec(1.0, bi, LambdaInv_i, 0.0);
    double auxf_new = -0.5 * count * VecVec(bi, b_lambda_inv)
  			      + VecVec(b_lambda_inv, BiasStats);
			      
    double auxf_change = auxf_new - auxf_old;

    model->b_.Row(i).CopyFromVec(bi);
    impr_vec(i) = auxf_change;
    if (i < 10 || auxf_change / (num_count+den_count+1.0e-10) > 2.0) {
      KALDI_LOG << "Updating bias: Auxf change per frame for Gaussian "
                << i << " is " << (auxf_change / num_count) << " over "
                << num_count << " frames " << "(den count was " << den_count;
    }
  }
  KALDI_VLOG(1) << "Updating b: numerator count is " << num_count_vec;
  KALDI_VLOG(1) << "Updating b: denominator count is " << den_count_vec;
  KALDI_VLOG(1) << "Updating b: objf-impr is " << impr_vec;
  
  double tot_count = num_count_vec.Sum(), tot_impr = impr_vec.Sum();
  tot_impr /= tot_count+1.0e-20;
  KALDI_LOG << "**Overall auxf impr for bias is " << tot_impr
            << " over " << tot_count << " frames";
  return tot_impr;
}

double EbwAmPldaUpdater::UpdateSubstateWeights( const MleAmPldaAccs &num_accs,
						      const MleAmPldaAccs &den_accs,
						      AmPlda *model) {
  KALDI_LOG << "Updating substate weights";

  double tot_count = 0.0, tot_impr_c = 0.0;
  for (int32 j = 0; j < num_accs.num_states_; j++) {
    int32 M = model->NumSubstates(j);
    Vector<double> num_occs(M), den_occs(M),
        orig_c_j(model->c_[j]), c_j(model->c_[j]);

    for (int32 m = 0; m < M; m++) {
      num_occs(m) = num_accs.gamma_[j].Row(m).Sum()
          + options_.tau_c * c_j(m);
      den_occs(m) = den_accs.gamma_[j].Row(m).Sum();
    }
    
    if (c_j.Dim() > 1) {
      double begin_auxf_c = 0.0, end_auxf_c = 0.0;
      for (int32 m = 0; m < M; m++) {  // see eq. 4.32, Dan Povey's PhD thesis.
        begin_auxf_c += num_occs(m) * log (c_j(m))
            - den_occs(m) * c_j(m) / orig_c_j(m);
      }
      for (int32 iter = 0; iter < 50; iter++) {
        Vector<double> k_jm(M);
        double max_m = 0.0;
        for (int32 m = 0; m < M; m++)
          max_m = std::max(max_m, den_occs(m)/orig_c_j(m));
        for (int32 m = 0; m < M; m++)
          k_jm(m) = max_m - den_occs(m)/orig_c_j(m);
        for (int32 m = 0; m < M; m++)
          c_j(m) = num_occs(m) + k_jm(m)*c_j(m);
        c_j.Scale(1.0 / c_j.Sum());
      }
      for (int32 m = 0; m < M; m++)
        c_j(m) = std::max(c_j(m),
                              static_cast<double>(options_.min_weight));
      c_j.Scale(1.0 / c_j.Sum()); // renormalize.

      for (int32 m = 0; m < M; m++) {
        end_auxf_c += num_occs(m) * log (c_j(m))
            - den_occs(m) * c_j(m) / orig_c_j(m);
      }
      tot_impr_c += end_auxf_c - begin_auxf_c;
      double this_impr_c = ((end_auxf_c - begin_auxf_c) / num_occs.Sum());
      if (j < 10 || this_impr_c > 0.5) {
        KALDI_LOG << "Updating substate weights: auxf impr for state " << j
                  << " is " << this_impr_c<< " per frame over " << num_occs.Sum()
                  << " frames (den count is " << den_occs.Sum() << ")";
      }
   
    }
    model->c_[j].CopyFromVec(c_j);
    tot_count += den_occs.Sum(); // Note: num and den occs should be the
    // same, except num occs are smoothed, so this is what we want.
  }    

  tot_impr_c /= (tot_count + 1.0e-20);

  KALDI_LOG << "**Overall auxf impr for c is " << tot_impr_c
            << " over " << tot_count << " frames";
  return tot_impr_c;
}  

double EbwAmPldaUpdater::UpdateComponentWeights( const MleAmPldaAccs &num_accs,
						      const MleAmPldaAccs &den_accs,
						      AmPlda *model) {
  KALDI_LOG << "Updating component weights";

  double tot_count = 0.0, tot_impr_w = 0.0;
  for (int32 j = 0; j < num_accs.num_states_; j++) {
    int32 num_gauss = model->NumGauss();
    Vector<double> num_occs(num_gauss), den_occs(num_gauss),
        orig_w_j(model->w_.Row(j)), w_j(model->w_.Row(j));

    for (int32 i = 0; i < num_gauss; i++) 
      for (int32 m = 0; m < model->NumSubstates(j); m++) {
	num_occs(i) = num_accs.gamma_[j](m,i)
          + options_.tau_w * w_j(i);
      den_occs(i) = den_accs.gamma_[j](m,i);
    }
    
    double begin_auxf_w = 0.0, end_auxf_w = 0.0;
    for (int32 i = 0; i < num_gauss; i++) {  // see eq. 4.32, Dan Povey's PhD thesis.
      begin_auxf_w += num_occs(i) * log (w_j(i))
          - den_occs(i) * w_j(i) / orig_w_j(i);
    }
    for (int32 iter = 0; iter < 50; iter++) {
      Vector<double> k_ji(num_gauss);
      double max_i = 0.0;
        for (int32 i = 0; i < num_gauss; i++)
          max_i = std::max(max_i, den_occs(i)/orig_w_j(i));
        for (int32 i = 0; i < num_gauss; i++)
          k_ji(i) = max_i - den_occs(i)/orig_w_j(i);
        for (int32 i = 0; i < num_gauss; i++)
          w_j(i) = num_occs(i) + k_ji(i)*w_j(i);
        w_j.Scale(1.0 / w_j.Sum());
      }
      for (int32 i = 0; i < num_gauss; i++)
        w_j(i) = std::max(w_j(i),
                              static_cast<double>(options_.min_weight));
      w_j.Scale(1.0 / w_j.Sum()); // renormalize.

      for (int32 i = 0; i < num_gauss; i++) {
        end_auxf_w += num_occs(i) * log (w_j(i))
            - den_occs(i) * w_j(i) / orig_w_j(i);
      }
      tot_impr_w += end_auxf_w - begin_auxf_w;
      double this_impr_w = ((end_auxf_w- begin_auxf_w) / num_occs.Sum());
      if (j < 10 || this_impr_w > 0.5) {
        KALDI_LOG << "Updating component weights: auxf impr for state " << j
                  << " is " << this_impr_w<< " per frame over " << num_occs.Sum()
                  << " frames (den count is " << den_occs.Sum() << ")";
      }
    
    model->w_.Row(j).CopyFromVec(w_j);
    tot_count += den_occs.Sum(); // Note: num and den occs should be the
    // same, except num occs are smoothed, so this is what we want.
  }
  
  tot_impr_w /= (tot_count + 1.0e-20);

  KALDI_LOG << "**Overall auxf impr for w is " << tot_impr_w
            << " over " << tot_count << " frames";
  return tot_impr_w;
    
}

}  // namespace kaldi
