// plda/ebw-am-plda.h

// Liang Lu (2014)

// See ../../COPYING for clarification regarding multiple authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.

#ifndef KALDI_PLDA_EBW_AM_PLDA_H_
#define KALDI_PLDA_EBW_AM_PLDA_H_ 1

#include <string>
#include <vector>

#include "gmm/model-common.h"
#include "itf/options-itf.h"
#include "plda/mle-am-plda.h"

namespace kaldi {

/**
   This header implements a form of Extended Baum-Welch training for SGMMs.
   If you are confused by this comment, see Dan Povey's thesis for an explanation of
   Extended Baum-Welch.
   A note on the EBW (Extended Baum-Welch) updates for the SGMMs... In general there is
   a parameter-specific value D that is similar to the D in EBW for GMMs.  The value of
   D is generally set to:
     E * (denominator-count for that parameter)   +   tau-value for that parameter
   where the tau-values are user-specified parameters that are specific to the type of
   the parameter (e.g. phonetic vector, subspace projection, etc.).  Things are a bit
   more complex for this update than for GMMs, because it's not just a question of picking
   a tau-value for smoothing: there is sometimes a scatter-matrix of some kind (e.g.
   an outer product of vectors, or something) that defines a quadratic objective function
   that we'll add as smoothing.  We have to pick where to get this scatter-matrix from.
   We feel that it's appropriate for the "E" part of the D to get its scatter-matrix from
   denominator stats, and the tau part of the D to get half its scatter-matrix from the
   both the numerator and denominator stats, assigned a weight proportional to how much
   stats there were.  When you see the auxiliary function written out, it's clear why this
   makes sense.

 */

struct EbwAmPldaOptions {
  BaseFloat tau_v; ///<  Smoothing constant for updates of sub-state vectors v_{jm}
  BaseFloat lrate_v; ///< Learning rate used in updating v-- default 0.5
  BaseFloat tau_G; ///<  Smoothing constant for the M quantities (phone-subspace projections)
  BaseFloat lrate_G; ///< Learning rate used in updating M-- default 0.5
  BaseFloat tau_U; ///<  Smoothing constant for the N quantities (speaker-subspace projections)
  BaseFloat lrate_U; ///< Learning rate used in updating N-- default 0.5
  BaseFloat tau_c;  ///< Tau value for smoothing substate weights (c)
  BaseFloat tau_w;  ///< Tau value for smoothing component weights  (w)
  BaseFloat tau_b;  ///< Tau value for smoothing component bias  (b)
  BaseFloat tau_var; ///< Tau value for smoothing covariance-matrices Lambda.
  BaseFloat lrate_var; ///< Learning rate used in updating Lambda-- default 0.5
  BaseFloat lrate_b;  ///< Learning rate used in updating bias 
  BaseFloat min_weight; ///< Minimum allowed weight in a sub-state.

  BaseFloat cov_min_value; ///< E.g. 0.5-- the maximum any eigenvalue of a covariance
  /// is allowed to change.  [this is the minimum; the maximum is the inverse of this,
  /// i.e. 2.0 in this case.  For example, 0.9 would constrain the covariance quite tightly,
  /// 0.1 would be a loose setting.
  
  BaseFloat max_cond; ///< large value used in SolveQuadraticProblem.
  BaseFloat epsilon;  ///< very small value used in SolveQuadraticProblem; workaround
  /// for an issue in some implementations of SVD.
  
  EbwAmPldaOptions() {
    tau_v = 50.0;
    lrate_v = 0.5;
    tau_G = 500.0;
    lrate_G = 0.5;
    tau_U = 500.0;
    lrate_U = 0.5;
    tau_c = 10.0;
    tau_w = 5.0;
    tau_b = 500.0;
    tau_var = 500.0;
    lrate_var = 0.5;
    lrate_b = 0.5;
    
    min_weight = 1.0e-05;
    cov_min_value = 0.5;
    
    max_cond = 1.0e+05;
    epsilon = 1.0e-40;
  }

  void Register(OptionsItf *po) {
    std::string module = "EbwAmPldaOptions: ";
    po->Register("tau-v", &tau_v, module+
                 "Smoothing constant for phone vector estimation.");
    po->Register("lrate-v", &lrate_v, module+
                 "Learning rate constant for phone vector estimation.");
    po->Register("tau-G", &tau_G, module+
                 "Smoothing constant for estimation of phonetic-subspace projections (G).");
    po->Register("lrate-G", &lrate_G, module+
                 "Learning rate constant for phonetic-subspace projections.");
    po->Register("tau-U", &tau_U, module+
                 "Smoothing constant for estimation of channel subspace projections (U).");
    po->Register("lrate-U", &lrate_U, module+
                 "Learning rate constant for channel-subspace projections.");
    po->Register("tau-c", &tau_c, module+
                 "Smoothing constant for estimation of substate weights (c)");
    po->Register("tau-w", &tau_w, module+
                 "Smoothing constant for estimation of component weights (w)");
    po->Register("tau-b", &tau_b, module+
                 "Smoothing constant for estimation of component bias (b)");
    po->Register("tau-var", &tau_var, module+
                 "Smoothing constant for estimation of within-class covariances (Sigma)");
    po->Register("lrate-var", &lrate_var, module+
                 "Constant that controls speed of learning for variances (larger->slower)");
    po->Register("lrate-b", &lrate_b, module+
                 "Constant that controls speed of learning for bias (larger->slower)");
    po->Register("cov-min-value", &cov_min_value, module+
                 "Minimum value that an eigenvalue of the updated covariance matrix can take, "
                 "relative to its old value (maximum is inverse of this.)");
    po->Register("min-weight", &min_weight, module+
                 "Floor for weights of sub-states and components.");
    po->Register("max-cond", &max_cond, module+
                 "Value used in handling singular matrices during update.");
    po->Register("epsilon", &epsilon, module+
                 "Value used in handling singular matrices during update.");
  }
};


/** \class EbwAmPldaUpdater
 *  Contains the functions needed to update the PLDA parameters.
 */
class EbwAmPldaUpdater {
 public:
  explicit EbwAmPldaUpdater(const EbwAmPldaOptions &options):
      options_(options) {}
  
  void Update(const MleAmPldaAccs &num_accs,
              const MleAmPldaAccs &den_accs,
              AmPlda *model,
              PldaUpdateFlagsType flags,
              BaseFloat *auxf_change_out,
              BaseFloat *count_out);
    
 protected:
  // The following two classes relate to multi-core parallelization of some
  // phases of the update.
//  friend class EbwUpdateWParallelClass;
//  friend class EbwUpdatePhoneVectorsClass;
 private:
  EbwAmPldaOptions options_;

  Vector<double> gamma_j_;  ///< State occupancies

  double UpdatePhoneVectors(const MleAmPldaAccs &num_accs,
                            const MleAmPldaAccs &den_accs,
                            AmPlda *model,
                            const std::vector< SpMatrix<double> > &H) const;
                                    
  double UpdateG(const MleAmPldaAccs &num_accs,
                 const MleAmPldaAccs &den_accs,
                 const std::vector< SpMatrix<double> > &Q_num,
                 const std::vector< SpMatrix<double> > &Q_den,
                 AmPlda *model) const;
  
  double UpdateU(const MleAmPldaAccs &num_accs,
                 const MleAmPldaAccs &den_accs,
                 AmPlda *model) const;
  
  double UpdateVars(const MleAmPldaAccs &num_accs,
                    const MleAmPldaAccs &den_accs,
                    AmPlda *model) const;
  
  double UpdateBias(const MleAmPldaAccs &num_accs,
                    const MleAmPldaAccs &den_accs,
                    AmPlda *model) const;
                                    
  double UpdateSubstateWeights(const MleAmPldaAccs &num_accs,
                               const MleAmPldaAccs &den_accs,
                               AmPlda *model);
			      
   double UpdateComponentWeights(const MleAmPldaAccs &num_accs,
                               const MleAmPldaAccs &den_accs,
                               AmPlda *model);

  KALDI_DISALLOW_COPY_AND_ASSIGN(EbwAmPldaUpdater);
  EbwAmPldaUpdater() {}  // Prevent unconfigured updater.
};


}  // namespace kaldi


#endif  // KALDI_PLDA_EBW_AM_PLDA_H_
