

#include <algorithm>
#include <limits>
#include <string>
#include <vector>

#include "gmm/diag-gmm.h"
#include "gmm/diag-gmm-normal.h"
#include "gmm/full-gmm.h"
#include "gmm/full-gmm-normal.h"
#include "plda/fa-gmm.h"


namespace kaldi {
  
  void FaGmm::Write(std::ostream &out_stream, bool binary) const {
  uint32 tmp_uint32;

  if (!valid_gconsts_)
    KALDI_ERR << "Must call ComputeGconsts() before writing the model.";
  
  WriteToken(out_stream, binary, "<FaGMM>");
  WriteToken(out_stream, binary, "<NUMGaussians>");
  tmp_uint32 = static_cast<uint32>(num_gauss_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  if (!binary) out_stream << "\n";

  WriteToken(out_stream, binary, "<GCONSTS>");
  gconsts_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<WEIGHTS>");
  weights_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<b>");
  b_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<inv_vars>");
  inv_vars_.Write(out_stream, binary);
  
  if (U_.size() != 0) {
    WriteToken(out_stream, binary, "<U>");
    for (int32 i = 0; i < num_gauss_; i++) {
      U_[i].Write(out_stream, binary);
    }
  }
  
    if (H_.size() != 0) {
      WriteToken(out_stream, binary, "<H>");
      for (int32 i = 0; i < num_gauss_; i++) {
        H_[i].Write(out_stream, binary);
      }
   }
  
  if (inv_V_.size() != 0) {
    WriteToken(out_stream, binary, "<inv_V>");
    for (int32 i = 0; i < num_gauss_; i++) {
      inv_V_[i].Write(out_stream, binary);
    }
  }
  
  if (inv_Sigma_.size() != 0) {
    WriteToken(out_stream, binary, "<inv_Sigma>");
    for (int32 i = 0; i < num_gauss_; i++) {
      inv_Sigma_[i].Write(out_stream, binary);
    }
  }

   if (G_.size() != 0) {
    WriteToken(out_stream, binary, "<G>");
    for (int32 i = 0; i < num_gauss_; i++) {
      G_[i].Write(out_stream, binary);
    }
  }
  
  WriteToken(out_stream, binary, "</FaGMM>");
 }

 void FaGmm::Read(std::istream &in_stream, bool binary){
   uint32 tmp_uint32;
   std::string token;
   
   ExpectToken(in_stream, binary, "<FaGMM>");
   
   ExpectToken(in_stream, binary, "<NUMGaussians>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   num_gauss_ = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<GCONSTS>");
   gconsts_.Read(in_stream, binary);
   ExpectToken(in_stream, binary, "<WEIGHTS>");
   weights_.Read(in_stream, binary);
   ExpectToken(in_stream, binary, "<b>");
   b_.Read(in_stream, binary);
   ExpectToken(in_stream, binary, "<inv_vars>");
   inv_vars_.Read(in_stream, binary);
   
   ReadToken(in_stream, binary, &token);
   while (token != "</FaGMM>") {
      if (token == "<U>") {
        U_.resize(num_gauss_);
        for (size_t i = 0; i < U_.size(); i++) {
           U_[i].Read(in_stream, binary);
        }
      } else if (token == "<H>"){
	H_.resize(num_gauss_);
	for (size_t i = 0; i < H_.size(); i++) {
           H_[i].Read(in_stream, binary);
	}
      } else if (token == "<inv_V>") {
        inv_V_.resize(num_gauss_);
        for (size_t i = 0; i < inv_V_.size(); i++) {
          inv_V_[i].Read(in_stream, binary);
        }
      } else if (token == "<inv_Sigma>") {
	inv_Sigma_.resize(num_gauss_);
	for (size_t i = 0; i< inv_Sigma_.size(); i++) {
	  inv_Sigma_[i].Read(in_stream, binary);
	}
      } else if (token == "<G>") {
	 G_.resize(num_gauss_);
        for (size_t i = 0; i < G_.size(); i++) {
           G_[i].Read(in_stream, binary);
        }
      }
      
      ReadToken(in_stream, binary, &token);
   }
   
  // ComputeSigma();
   ComputeVarOffset();
   ComputeGconsts();  // safer option than trusting the read gconsts
 }

  void FaGmm::RandU() {
        U_.resize(num_gauss_);
        for(int32 m=0; m<NumGauss(); m++) {
          U_[m].SetRandn();
        }
        ComputeV();
        ComputeH();
        ComputeSigma();
        ComputeGconsts();
  }

  void FaGmm::InitializeG(int32 state_dim) {
    
    if (G_.size() == 0) {
      G_.resize(num_gauss_);
      for(int32 m=0; m<NumGauss(); m++) {
	G_[m].Resize(U_[m].NumRows(), state_dim);
        G_[m].CopyFromMat(U_[m].ColRange(0, state_dim));  
//	U_[m].SetZero();
      }
    }
  }
  
  void FaGmm::ComputeVarOffset() {

  SpMatrix<double> tmp_mat(FeatureDim());
  var_offset_.Resize(NumGauss(), FeatureDim());
  for(int m=0; m<NumGauss(); m++) {
     tmp_mat.AddMat2Sp(1.0, U_[m], kNoTrans, inv_V_[m], 0.0);
     var_offset_.Row(m).CopyDiagFromSp(tmp_mat);  
  }
}

  //initialize FaGmm from a diagonal-covariance GMM 
  void FaGmm::InitFromDiagGmm(const DiagGmm &gmm, int32 speech_dim, int32 state_dim) {
   
    num_gauss_ = gmm.NumGauss();
    int32 feature_dim = gmm.Dim();
    KALDI_ASSERT(speech_dim > 0 && 
			num_gauss_ > 0 && 
			feature_dim >= speech_dim &&
			speech_dim >= state_dim);
    
    Vector<double> mean_vec(feature_dim);
    Vector<double> tmp_vec(feature_dim);
    SpMatrix<double> global_var(feature_dim);
    Vector<double> global_mean(feature_dim);
    Vector<double> eigenvals(speech_dim);
    Matrix<double> eigenvecs(feature_dim, speech_dim);
    
    // speech_dim_ = dim;
    weights_.Resize(num_gauss_);
    inv_vars_.Resize(num_gauss_, feature_dim);
     b_.Resize(num_gauss_, feature_dim);
    weights_.CopyFromVec(gmm.weights());
    inv_vars_.CopyFromMat(gmm.inv_vars());
//    gmm.GetVars(&inv_vars_);
    gmm.GetMeans(&b_);
    
    //initialize the offset means
    global_var.SetZero();
    global_mean.SetZero();
    for(int i=0; i<num_gauss_; i++) {
      mean_vec.CopyRowFromMat(b_, i);
      global_var.AddVec2(1.0, mean_vec);  //collecting second-order statistics
      global_mean.AddVec(1.0, mean_vec); //collecting first-order statistics
    }//for
    
    //initialize the projection matrices
    U_.resize(num_gauss_);
    global_mean.Scale(1.0/num_gauss_);
    global_var.Scale(1.0/num_gauss_);
    global_var.AddVec2(-1.0, global_mean);
    global_var.TopEigs(&eigenvals, &eigenvecs); //do svd
  //  eigenvals.ApplyPow(0.5);
  //  eigenvals.Scale(0.1);  //use smaller eigenvalues to initialize U
//    eigenvecs.MulColsVec(eigenvals);
    
    KALDI_LOG<<"eigen values ="<<eigenvals;
    
    for(int m=0; m<num_gauss_; m++) {
      U_[m].Resize(feature_dim, speech_dim);
      U_[m].CopyFromMat(eigenvecs);    //the projection matrices are initialized to be the same
    }
    
    G_.resize(num_gauss_);
    for(int32 m=0; m<num_gauss_; m++) {
      G_[m].Resize(feature_dim, state_dim);
      G_[m].CopyFromMat(U_[m].ColRange(0, state_dim));  
    }
    
    inv_Sigma_.resize(num_gauss_);
    for(int m=0; m<num_gauss_; m++) {
      inv_Sigma_[m].Resize(feature_dim);
      inv_Sigma_[m].SetZero();
      inv_Sigma_[m].AddVec(1.0, inv_vars_.Row(m));
    }
 
    ComputeV();
    ComputeH();
    ComputeGconsts();
  }

  //initialize FaGmm from a full-covariance GMM 
  void FaGmm::InitFromFullGmm(const FullGmm &gmm, int32 speech_dim) {
  
 //   num_gauss_ = gmm.NumGauss();
    num_gauss_ = 1;  //just use one Gauss
    int32 feature_dim = gmm.Dim();
    KALDI_ASSERT(speech_dim > 0 && 
			num_gauss_ > 0 && 
			feature_dim >= speech_dim);
    
    Matrix<double> means(gmm.NumGauss(), feature_dim);
    Vector<double> tmp_vec(feature_dim);
    std::vector< SpMatrix<double> > covars;
    Vector<double> eigenvals(speech_dim);
    Matrix<double> eigenvecs(feature_dim, speech_dim);
    
    // speech_dim_ = dim;
    weights_.Resize(num_gauss_);
    inv_vars_.Resize(num_gauss_, feature_dim);
 //   weights_.CopyFromVec(gmm.weights());
    weights_(0) = 1.0;
    gmm.GetMeans(&means); 		//initialize the offset means
    gmm.GetCovars(&covars);
    
    b_.Resize(num_gauss_, feature_dim);
    b_.Row(0).CopyRowFromMat(means, 0);
    
    //initialize the inv_Sigma_
    inv_Sigma_.resize(num_gauss_);
    for(int m=0; m<num_gauss_; m++) {
      inv_Sigma_[m].Resize(feature_dim);
      inv_Sigma_[m].CopyFromSp(covars[m]);
      inv_Sigma_[m].InvertDouble();
    }
    
    //initialize the projection matrices
    U_.resize(num_gauss_);
    for(int i=0; i<num_gauss_; i++) {
       inv_vars_.Row(i).CopyDiagFromSp(covars[i]);  //initialize inv_vars_
       inv_vars_.Row(i).InvertElements();
       covars[i].TopEigs(&eigenvals, &eigenvecs); //do svd
       eigenvals.ApplyPow(0.5);
       eigenvals.Scale(0.1); //use smaller eigenvalues to initialize U
       eigenvecs.MulColsVec(eigenvals);
       U_[i].Resize(feature_dim, speech_dim);
       U_[i].CopyFromMat(eigenvecs);   
       KALDI_LOG<<"eigen values ="<<eigenvals;
    }//for
 
    ComputeV();
    ComputeH();
    ComputeGconsts();
    
  }
  
  ///NOTE: V is stored as its inverse
  void FaGmm::ComputeV() {
    
    int32 speech_dim = SpeechDim();
    int32 feature_dim = FeatureDim();
    KALDI_ASSERT(num_gauss_ > 0 && 
			    speech_dim > 0 &&
			    feature_dim > 0);
    
    Matrix<double> I(speech_dim, speech_dim);
    Vector<double> inv_var(feature_dim);
    
    I.SetUnit();
    inv_V_.resize(num_gauss_);
    for(int m=0; m<num_gauss_; m++) {
      inv_var.CopyRowFromMat(inv_vars_, m);
      inv_V_[m].Resize(speech_dim);
      inv_V_[m].CopyFromMat(I);
      inv_V_[m].AddMat2Vec(1.0, U_[m], kTrans, inv_var, 1.0);  //V(m) = I + U(m)^T*\Lambda(m)^(-1)*U
      inv_V_[m].InvertDouble();
    }//for
    
  }
  
  void FaGmm::ComputeH() {
    int32 speech_dim = SpeechDim();
    int32 feature_dim = FeatureDim();
    KALDI_ASSERT(num_gauss_ > 0 && 
			    speech_dim > 0 &&
			    feature_dim > 0);
			    
    H_.resize(num_gauss_);
    for(int m=0; m<num_gauss_; m++) {
      H_[m].Resize(speech_dim, feature_dim);
      H_[m].AddSpMat(1.0, inv_V_[m], U_[m], kTrans, 0.0);
      H_[m].MulColsVec(inv_vars_.Row(m));
    }
    
  }
  
  ///\Sigma is the covariance matrix of observations p(y)
  ///e.g. \Sigma_[m] = \Lambda_m + U_mU_m^T
  ///NOTE: we store the inverted version
  void FaGmm::ComputeSigma() {
    
    int32 feature_dim = FeatureDim();
    KALDI_ASSERT(num_gauss_ > 0 && feature_dim > 0);
    Vector<double> inv_var(feature_dim);
    
    if(inv_Sigma_.size() != num_gauss_) 
      inv_Sigma_.resize(num_gauss_);
    
    for(int m=0; m<num_gauss_; m++) {
      inv_var.CopyRowFromMat(inv_vars_, m);
      inv_var.InvertElements();
      inv_Sigma_[m].Resize(feature_dim);
      inv_Sigma_[m].SetZero();
      inv_Sigma_[m].AddVec(1.0, inv_var);
      inv_Sigma_[m].AddMat2(1.0, U_[m], kNoTrans, 1.0);
      inv_Sigma_[m].InvertDouble();
    }
    
    //since \Sigma change, recompute gconsts
    ComputeGconsts();
  }
  
  ///Now we have introduced the class-based subspace,
  ///so we update the Sigma, \Sigma_[m] = \Lambda_m + U_mU_m^T + G_mG_m^T
  ///NOTE: we store the inverted version
  void FaGmm::UpdateSigmaPLDA(){
    
    int32 feature_dim = FeatureDim();
    KALDI_ASSERT(num_gauss_ > 0 && feature_dim > 0);
    Vector<double> inv_var(feature_dim);
    
    if(inv_Sigma_.size() != num_gauss_) 
      inv_Sigma_.resize(num_gauss_);
    
    for(int m=0; m<num_gauss_; m++) {
      inv_var.CopyRowFromMat(inv_vars_, m);
      inv_var.InvertElements();
      inv_Sigma_[m].Resize(feature_dim);
      inv_Sigma_[m].SetZero();
      inv_Sigma_[m].AddVec(1.0, inv_var);
      inv_Sigma_[m].AddMat2(1.0, U_[m], kNoTrans, 1.0);
      inv_Sigma_[m].AddMat2(1.0, G_[m], kNoTrans, 1.0);
      inv_Sigma_[m].InvertDouble();
    }
    
    //since \Sigma change, recompute gconsts
    ComputeGconsts();
  }
  
  
  Vector<double> 
  FaGmm::CondPosteriorMean(Vector<double> const &data, int32 m) const {
   
    KALDI_ASSERT(data.Dim() == FeatureDim());
    
    int32 feature_dim = FeatureDim();
    int32 speech_dim = SpeechDim();
    Vector<double> hat_xt(speech_dim);
    Vector<double> tmp_vec(feature_dim);
    
    tmp_vec.CopyFromVec(data);
    tmp_vec.AddVec(-1.0, b_.Row(m));
    
//    inv_var.CopyRowFromMat(inv_vars_, m);
//    tmp_vec.MulElements(inv_var);
    
//    w_mt.AddMatVec(1.0, U_[m], kTrans, tmp_vec, 0.0); //w_mt = U_m^T*\Lambda_m^(-1)*(\bar(y_t) - b_m)
//    hat_xt.AddSpVec(1.0, inv_V_[m], w_mt, 0.0);  //hat_xt = V_m^{-1}*w_mt

    hat_xt.AddMatVec(1.0, H_[m], kNoTrans, tmp_vec, 0.0); 
    return hat_xt;
  }
  
  ///Given the state vector, estimate the mean p(x|y, z, m)
  Vector<double> 
  FaGmm::PosteriorFrameMeanPLDA(Vector<double> const &data, 
						      Vector<double> const &z_mean,
						      int32 m) const {
   
    KALDI_ASSERT(data.Dim() == FeatureDim());
    
    int32 feature_dim = FeatureDim();
    int32 speech_dim = SpeechDim();
    Vector<double> hat_xt(speech_dim);
    Vector<double> tmp_vec(feature_dim);
    
    tmp_vec.CopyFromVec(data);
    tmp_vec.AddVec(-1.0, b_.Row(m));
    tmp_vec.AddMatVec(-1.0, G_[m], kNoTrans, z_mean, 1.0);  //remove the state-based bias
//    inv_var.CopyRowFromMat(inv_vars_, m);
//    tmp_vec.MulElements(inv_var);
    
//    w_mt.AddMatVec(1.0, U_[m], kTrans, tmp_vec, 0.0); //w_mt = U_m^T*\Lambda_m^(-1)*(\bar(y_t) - b_m)
//    hat_xt.AddSpVec(1.0, inv_V_[m], w_mt, 0.0);  //hat_xt = V_m^{-1}*w_mt

    hat_xt.AddMatVec(1.0, H_[m], kNoTrans, tmp_vec, 0.0); 
    return hat_xt;
  }
  
  Vector<double>
  FaGmm::AvgCondPosteriorMean(Vector<double> const &data,
				const std::vector<int32> &gselect) {
    Vector<double> posterior;
    Vector<double> hat_xt(SpeechDim());
    double loglike;
  
    hat_xt.SetZero();
    loglike = ComponentPosteriors(data, gselect, &posterior);
    for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];
      double this_posterior = posterior(ki);
      if(this_posterior > 0) {
        Vector<double> x_mean = CondPosteriorMean(data, m);
 //     KALDI_LOG<<"m = "<<m<<" xt = "<<x_mean;
        hat_xt.AddVec(this_posterior, x_mean);
      }
    }
    
    return hat_xt;
  }

  BaseFloat FaGmm::CondPosteriorsAndMeansPLDA(Vector<double> const &data,
				const std::vector<int32> &gselect,
				Matrix<double> const &z_means,
				Matrix<BaseFloat> *x_means,
				Vector<BaseFloat> *posteriors) { 
    double loglike;
    Vector<double> posteriors_d;
    Vector<double> tmp_z(z_means.NumCols());
    x_means->Resize(gselect.size(), SpeechDim());
    loglike = ComponentPosteriors(data, gselect, &posteriors_d);
    
    posteriors->Resize(gselect.size());
    posteriors->CopyFromVec(posteriors_d);
    
    for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];    
      tmp_z.CopyRowFromMat(z_means, m);
      Vector<BaseFloat> hat_xt(PosteriorFrameMeanPLDA(data, tmp_z, m));
 //   KALDI_LOG<<"m = "<<m<<" xt = "<<x_mean;
      x_means->CopyRowFromVec(hat_xt, ki);
      
    }
    
    return loglike;
  }
  
  BaseFloat FaGmm::CondPosteriorsAndMeans(Vector<double> const &data,
				const std::vector<int32> &gselect,
				Matrix<BaseFloat> *x_means,
				Vector<BaseFloat> *posteriors) { 
    double loglike;
    Vector<double> posteriors_d;
    x_means->Resize(gselect.size(), SpeechDim());
    loglike = ComponentPosteriors(data, gselect, &posteriors_d);
    
    posteriors->Resize(gselect.size());
    posteriors->CopyFromVec(posteriors_d);
    
    for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];    
      Vector<BaseFloat> hat_xt(CondPosteriorMean(data, m));
 //   KALDI_LOG<<"m = "<<m<<" xt = "<<x_mean;
      x_means->CopyRowFromVec(hat_xt, ki);
      
    }
    
    return loglike;
  }
  
//compute the loglikelhood p(y|m) 
  double FaGmm::ComponentMargLoglik(Vector<double> const &data, 
								    int32 m) {
    if(!valid_gconsts_)  ComputeGconsts();
    
    double loglike = gconsts_(m);
    Vector<double> y_hat(data);
    Vector<double> bias(b_.NumCols());
    
    bias.CopyRowFromMat(b_, m);
    y_hat.AddVec(-1.0, bias);
    
    loglike -= 0.5 * VecSpVec(y_hat, inv_Sigma_[m], y_hat);
    
    return loglike;
  }

//compute the loglikelhood p(y) by marginalize out x
  double FaGmm::MargLogLikelihood(Vector<double> const &data, 
			      const std::vector<int32> &gselect) {
    
    Vector<double> logp_y(gselect.size());
    
    for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];
      logp_y(m) = ComponentMargLoglik(data, m);
    }
    
    return logp_y.LogSumExp();
  }
  
  int32 FaGmm::ComputeGconsts() {
    
    int32 num_gauss = NumGauss();
    int32 dim = FeatureDim();
    double offset = -0.5 * M_LOG_2PI * dim;  // constant term in gconst.
    int32 num_bad = 0;

    // Resize if Gaussians have been removed during Update()
    if (num_gauss != gconsts_.Dim())
      gconsts_.Resize(num_gauss);

    for (int32 mix = 0; mix < num_gauss; mix++) {
      KALDI_ASSERT(weights_(mix) >= 0);  // Cannot have negative weights.
      double gc = log(weights_(mix)) + offset;  // May be -inf if weights == 0
      gc += 0.5 * inv_Sigma_[mix].LogPosDefDet();  //change sign because it's inverse covariance
      
      if (KALDI_ISNAN(gc)) {  // negative infinity is OK but NaN is not acceptable
        KALDI_ERR << "At component "  << mix
                << ", not a number in gconst computation";
      }
      if (KALDI_ISINF(gc)) {
        num_bad++;
        // If positive infinity, make it negative infinity.
        // Want to make sure the answer becomes -inf in the end, not NaN.
        if (gc > 0) gc = -gc;
      }
      gconsts_(mix) = gc;
    }

    valid_gconsts_ = true;
    return num_bad;
  }
  
  double FaGmm::ComponentPosteriors(Vector<double> const &data,
							const std::vector<int32> &gselect,
							Vector<double> *post) {
							  
     if (post == NULL) KALDI_ERR << "NULL pointer passed as return argument.";
     int32 num_gselect = gselect.size();
     post->Resize(num_gselect);							  
     
     for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];
      (*post)(ki) = ComponentMargLoglik(data, m);
//      KALDI_LOG<<"m = "<<m<<" ki = "<<ki<<" post[m] = "<<(*post)(ki);
    }
   
//    if(kapa > 0) post->Scale(1.0/kapa);
    double loglike = post->ApplySoftMax();
/*
    for(int32 i=0; i<post->Dim(); i++) {
      if((*post)(i) < posterior_threshold) {
	(*post)(i) = 0.0;
      }
    }
  */  
    return loglike;
  }

  void FaGmm::ComputeLogDet() {
  
     inv_vars_LogDet_.Resize(num_gauss_);
     inv_V_LogDet_.Resize(num_gauss_);
     
     for(int32 m=0; m<num_gauss_; m++){
       inv_vars_LogDet_(m) = inv_vars_.Row(m).SumLog();
       inv_V_LogDet_(m) = inv_V_[m].LogPosDefDet();
     }
  }
  
  void FaGmm::ComputeX() {
    X_.resize(num_gauss_);
    Matrix<double> U_trans(SpeechDim(), FeatureDim());
    for(int32 m=0; m<num_gauss_; m++){
      X_[m].Resize(SpeechDim(), StateDim());
      U_trans.CopyFromMat(U_[m], kTrans);
      U_trans.MulColsVec(inv_vars_.Row(m));
      X_[m].AddMatMat(1.0, U_trans, kNoTrans, G_[m], kNoTrans, 0.0);
    }
  }
  
  void FaGmm::CholeskyV() {
    inv_V_sqrt_.resize(num_gauss_);
    for (int32 m=0; m<num_gauss_; m++) {
      inv_V_sqrt_[m].Resize(SpeechDim());
      inv_V_sqrt_[m].Cholesky(inv_V_[m]);
    }
  }
  
  void FaGmm::ComputeTraceMat() {
    
    tr_mat_.Resize(num_gauss_, StateDim());
    Matrix<double> tmp_mat(SpeechDim(), StateDim());
    SpMatrix<double> tmp_sp(StateDim());
    Matrix<double> U_trans(SpeechDim(), FeatureDim());
    for(int32 m=0; m<num_gauss_; m++) {
      U_trans.CopyFromMat(U_[m], kTrans);
      U_trans.MulColsVec(inv_vars_.Row(m));
      tmp_mat.AddMatMat(1.0, U_trans, kNoTrans, G_[m], kNoTrans, 0.0);
      tmp_sp.AddMat2Sp(1.0, tmp_mat, kTrans, inv_V_[m], 0.0);
      tmp_sp.AddMat2Vec(1.0, G_[m], kTrans, inv_vars_.Row(m), 1.0);
      tr_mat_.Row(m).CopyDiagFromSp(tmp_sp);
    }
  }
  
}//end of namespace
  
  
