

#ifndef KALDI_FA_GMM_H_
#define KALDI_FA_GMM_H_ 1

#include <utility>
#include <vector>

#include "base/kaldi-common.h"
#include "gmm/model-common.h"
#include "matrix/matrix-lib.h"
#include "gmm/diag-gmm.h"
#include "gmm/full-gmm.h"


#ifndef posterior_threshold
#  define posterior_threshold 1.0e-2
#endif



namespace kaldi {
  
  class FaGmm {
    
    public:
      
      ///empty construction
      FaGmm():valid_gconsts_(false) {}
      
      explicit FaGmm(const FaGmm &fgmm): valid_gconsts_(false) {
         CopyFromFaGmm(fgmm);
      }
  
      ///IO functions
      void Read(std::istream &in_stream, bool binary);
      void Write(std::ostream &out_stream, bool binary) const;
  
      ///copy from a fa-gmm
      void CopyFromFaGmm(const FaGmm &fgmm);
      
      ///initialize from a full-covariance GMM
      void InitFromFullGmm(const FullGmm &gmm, int32 speech_dim);
      
      ///initialize from a diagonal-covariance GMM
      void InitFromDiagGmm(const DiagGmm &gmm, int32 speech_dim,  int32 state_dim);
      
      ///initialize G for PLDA
      void InitializeG(int32 state_dim);
     
      /// initialize U with random matrices
      void RandU();
 
      ///return the conditional loglikelihood for a component, e.g. logp(y|x, m)
      double ComponentCondLoglik(Vector<double> const &data, 
							    Vector<double> &x,
							    int32 m) const;
    
      ///return the marginal loglikelihood for a component, e.g. logp(y| m)						    
      double ComponentMargLoglik(Vector<double> const &data, int32 m);
      
      ///return the total marginal loglikelihood
      double MargLogLikelihood(Vector<double> const &data,
						const std::vector<int32> &gselect);
      
      ///return the total conditional loglikelihood
      double CondLogLikelihood(Vector<double> const &data) const;
      
      ///compute the covariance matrix of the latent speech distribution: p(x|y,m)
      void ComputeV();
      
      ///compute the transform matrix H for estimating means of p(x|y,m)
      void ComputeH();
      
      ///compute the covariance matrix of observations p(y)
      void ComputeSigma();
      
      ///update covariance matrix for PLDA model
      void UpdateSigmaPLDA();
      
      ///compute the mean of the latent speech distribution given a data point: E[p(x|y,m)]
      Vector<double> CondPosteriorMean(Vector<double> const &data, int32 m) const;
      
      ///compute the mean of the latent speech distribution given a data point: E[p(x|y,z,m)]
      Vector<double> PosteriorFrameMeanPLDA(Vector<double> const &data, 
						      Vector<double> const &z_mean,
						      int32 m) const;
						      
      ///compute the average mean of the lattent speech distribution a given a data point: E[p(x|y)]
      Vector<double> AvgCondPosteriorMean(Vector<double> const &data,
									const std::vector<int32> &gselect);

       BaseFloat CondPosteriorsAndMeans(Vector<double> const &data,
				const std::vector<int32> &gselect,
				Matrix<BaseFloat> *x_means,
				Vector<BaseFloat> *posteriors);
       
       BaseFloat CondPosteriorsAndMeansPLDA(Vector<double> const &data,
				const std::vector<int32> &gselect,
				Matrix<double> const &z_means,
				Matrix<BaseFloat> *x_means,
				Vector<BaseFloat> *posteriors);
				
      ///compute the posteriors p(m|y) for selected components
      double ComponentPosteriors(Vector<double> const &data,
							const std::vector<int32> &gselect,
							Vector<double> *post);
      
      ///compute the values of gconsts_
      int32 ComputeGconsts();
      
      void ComputeLogDet();
      
      void ComputeX();
      
      ///do Cholesky decomposition of inv_V_
      void CholeskyV();
      
      ///compute the matrix for the trace term in the loglikelihood function
      void ComputeTraceMat();
      void ComputeVarOffset();
      
      ///Accessors
      int32 FeatureDim() const { return b_.NumCols(); }
      int32 SpeechDim() const { return U_[0].NumCols(); }
      int32 StateDim() const { return G_[0].NumCols(); }
      int32 NumGauss() const { return U_.size(); }
      
      Vector<double> gconsts() const {return gconsts_; }
      
      Vector<double> weights() const {return weights_; };
      
      Vector<double> InvCovars(int32 m) const {
	 KALDI_ASSERT(m<NumGauss());
	 Vector<double> var_m(FeatureDim());
	 var_m.CopyRowFromMat(inv_vars_, m);
	 return var_m;
      }
      
      Vector<double> BiasVector(int32 m) const {
	  KALDI_ASSERT(m<NumGauss());
	  Vector<double> b_m(FeatureDim());
	  b_m.CopyRowFromMat(b_, m);
	  return b_m;
      }
      
      Matrix<double> Projection(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return U_[m];
      }
      
       Matrix<double> PreXform(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return X_[m];
      }
      
      Matrix<double> StateProjection(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return G_[m];
      }
      
      SpMatrix<double> InverseV(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return inv_V_[m];
      }
      
      TpMatrix<double> InverseVsqrt(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return inv_V_sqrt_[m];
      }
      
      SpMatrix<double> Inv_Sigma(int32 m) const {
        KALDI_ASSERT(m<NumGauss());
	return inv_Sigma_[m];
      }
      
       double InverseVLogDet(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return inv_V_LogDet_(m);
      }
      
      double InvCovarsLogDet(int32 m) const {
	KALDI_ASSERT(m<NumGauss());
	return inv_vars_LogDet_(m);
      }
      
      Matrix<double> TraceMat() const {
	return tr_mat_;
      }
      
      const Matrix<double> & VarOffset() const {
        return var_offset_;  
      }
   
    private:
      
      ///back-ground GMM model
      DiagGmm diag_ubm_;
      FullGmm full_ubm_;
     
      ///number of Gaussian component
      int32 num_gauss_;
      
      ///vector of constant values to compute the probability p(y)
      ///e.g. gconst_(i) = -0.5(log(2pi) + log |Sigma_[i]| )
      Vector<double> gconsts_;
      bool valid_gconsts_;      ///recompute gconst_  if false
      ///vector of comonent weights
      Vector<double> weights_;
      ///projection matrix, dim [I][D][S]
      std::vector< Matrix<double> > U_;
      ///state-related projection matrix, dim[I][D][T]
      std::vector< Matrix<double> > G_;
      ///offset mean vectors, stacked in a matrix, dim [I][D]
      Matrix<double> b_;
      ///residual inverse diagonal covariance matrix, stacked in a matrix, dim [I][D]
      Matrix<double> inv_vars_;
      Vector<double> inv_vars_LogDet_;
      ///posterior per-component variance matrix, 
      ///NOTE there are not precision matrices, dim [I][S][S]
      std::vector< SpMatrix<double> > inv_V_; 
      ///Cholesky decomposition of inv_V_;
      std::vector< TpMatrix<double> > inv_V_sqrt_;
      ///LogDet of inv_V_
      Vector<double> inv_V_LogDet_;
      ///posterior per-state vector variance matrix,
      ///NOTE there are not precision matrices, dim[I][T][T]
      std::vector< SpMatrix<double> > inv_F_;
      ///transform matrix that used to rapid computing of the posterior mean, dim [I][S][D]
      std::vector< Matrix<double> > H_;
      ///inverse covariance matrices for speech observations p(y), dim[I][D][D]
      ///inv_Sigma_[m] = (\Lambda_m + U_m^T*U_m)^{-1}
      std::vector< SpMatrix<double> > inv_Sigma_;
      ///prexform U_m^T*inv_vars*G_m
      std::vector< Matrix<double> > X_;
      ///matrix to compute the trace term
      Matrix<double> tr_mat_;
      ///diag(UVInvU^T), used for covariance update
     Matrix<double> var_offset_;
   
      friend class MleFaGmmUpdater;
  };
  
}

#endif
