// plda/mle-am-plda.h

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.


#ifndef KALDI_PLDA_MLE_AM_PLDA_H_
#define KALDI_PLDA_MLE_AM_PLDA_H_ 1

#include "gmm/model-common.h"
#include "plda/am-plda.h"
#include "itf/options-itf.h"

namespace kaldi {
using std::string;
using std::vector;

/** \struct MleAmPldaOptions
 *  Configuration variables needed in the AmPlda estimation process.
 */
struct MleAmPldaOptions {
  /// Configuration Parameters.  See initialization code for more comments.
  BaseFloat tau_vec;  ///< Amount of smoothing for v_{jm} update
  BaseFloat tau_c;    ///< Tau value for smoothing substate weights (c)
  /// Floor covariance matrices Sigma_i to this times average cov.
  BaseFloat cov_floor;
  BaseFloat min_gaussian_weight;
  BaseFloat min_gaussian_occupancy;
  /// Max on condition of matrices in update beyond which we do not update.
  /// Should probably be related to numerical properties of machine
  /// or BaseFloat type.
  BaseFloat max_cond;
  /// Limits condition of smoothing matrices H_sm (e.g. 100).
  /// Only really important on 1st iter if using priors.
  BaseFloat max_cond_H_sm;
  /// Fix for the smoothing approach, necessary if max_cond_H_sm != inf
  /// note: only has an effect if tau_vec != 0.
//  bool fixup_H_sm;
  /// Set check_v to true if you want to use the "checking" version of the update
  /// for the v's, in which it checks the "real" objective function value and
  /// backtracks if necessary;
//  bool check_v;

  bool renormalize_V;  // Renormalize the phonetic space.

  BaseFloat epsilon;  ///< very small value used to prevent SVD crashing.

  MleAmPldaOptions() {
    // tau value used in smoothing vector re-estimation (if no prior used).
    tau_vec = 0.0;
    tau_c  = 5.0;
    cov_floor = 0.01;
    min_gaussian_weight = 1.0e-05;
    min_gaussian_occupancy = 5.0;
    max_cond = 1.0e+05;
    epsilon = 1.0e-40;
    max_cond_H_sm = 1.0e+05;  // only for diagnostics in normal situations.
  //  fixup_H_sm = true;
 //   check_v = false;  // for back-compat.
    renormalize_V = false;
  //  renormalize_N = false;  // default to false since will invalidate spk vectors
                            // on disk.
  }

  void Register(OptionsItf *po) {
    std::string module = "MleAmPldaOptions: ";
    po->Register("tau-vec", &tau_vec, module+
                 "Smoothing for phone vector estimation.");
    po->Register("tau-c", &tau_c, module+
                 "Smoothing for substate weights estimation.");
    po->Register("cov-floor", &cov_floor, module+
                 "Covariance floor (fraction of average covariance).");
    po->Register("max-cond", &max_cond, module+"Maximum condition number beyond"
                 " which matrices are not updated.");
//    po->Register("check-v", &check_v, module+"If true, check real auxf "
//                 "improvement in update of v and backtrack if needed "
 //                "(not compatible with smoothing v)");
    po->Register("min-gaussian-weight", &min_gaussian_weight,
                 module+"Min Gaussian weight before we remove it.");
    po->Register("min-gaussian-occupancy", &min_gaussian_occupancy,
                 module+"Minimum occupancy to update a Gaussian.");
    po->Register("renormalize-v", &renormalize_V, module+"If true, renormalize "
                 "the phonetic-subspace vectors to have meaningful sizes.");
  }
};

/** \class MleAmPldaAccs
 *  Class for the accumulators associated with the AmPlda parameters 
 */
class MleAmPldaAccs {
 public:
  explicit MleAmPldaAccs(BaseFloat rand_prune = 1.0e-05)
      : total_frames_(0.0), total_like_(0.0), feature_dim_(0),
        phn_dim_(0), chn_dim_(0), num_gauss_(0),
        num_states_(0), rand_prune_(rand_prune) {}

  MleAmPldaAccs(const AmPlda &model, PldaUpdateFlags flags,
                BaseFloat rand_prune = 1.0e-05)
      : total_frames_(0.0), total_like_(0.0), rand_prune_(rand_prune) {
    ResizeAccumulators(model, flags);
  }

  //~MleAmPldaAccs();

  void Read(std::istream &in_stream, bool binary, bool add);
  void Write(std::ostream &out_stream, bool binary) const;

  /// Checks the various accumulators for correct sizes given a model. With
  /// wrong sizes, assertion failure occurs. When the show_properties argument
  /// is set to true, dimensions and presence/absence of the various
  /// accumulators are printed. For use when accumulators are read from file.
  void Check(const AmPlda &model, bool show_properties = true) const;

  /// Resizes the accumulators to the correct sizes given the model. The flags
  /// argument control which accumulators to resize.
  void ResizeAccumulators(const AmPlda &model, PldaUpdateFlagsType flags);

  /// Returns likelihood.
  BaseFloat Accumulate(const AmPlda &model,
                       const PldaPerFrameDerivedVars &frame_vars,
                       int32 state_index, BaseFloat weight,
                       PldaUpdateFlagsType flags);

  /// Accumulate the stats given the per-Gaussian component posteriors 
  BaseFloat AccumulateFromPosteriors(const AmPlda &model,
                                     const PldaPerFrameDerivedVars &frame_vars,
                                     const Matrix<BaseFloat> &posteriors,
                                     int32 j,
                                     PldaUpdateFlagsType flags);
                                  
  /// Accessors
  void GetStateOccupancies(Vector<BaseFloat> *occs) const {
     occs->Resize(gamma_.size());
     for (int32 j = 0, end = gamma_.size(); j < end; j++) {
       (*occs)(j) = gamma_[j].Sum();
      }
  }
  
  const std::vector< Matrix<double> >& GetOccs() const {
    return gamma_;
  }
  
 private:
  /// Flags corresponding to the accumulators that are stored.
  PldaUpdateFlagsType flags_;

   ///First part of the stats to update U: sum_t w_{mt}(y_t - b_m)E^T[x_t]
   std::vector< Matrix<double> > U_accumulator_first_;  
   ///Second part of the stats to update U: sum_t w_{mt} E[x_t*x_t^T]
   std::vector< SpMatrix<double> > U_accumulator_second_;
   ///First part of the stats to update G
   std::vector< Matrix<double> > G_accumulator_;
 //  ///Second part of the stats to update G
 //  std::vector< SpMatrix<double> > G_accumulator_second_;
   ///Statistics to update bias vector b_m
   Matrix<double> b_accumulator_;
   ///Statistics to update the variance \Lambda_m
   Matrix<double> var_accumulator_;
   ///Statistics to update prior mean of x
   Matrix<double> mean_x_acc_;
   ///Statistics to update prior covariance of x
   std::vector< SpMatrix<double> > cov_x_acc_;
   
  /// The AmPlda state specific stats.
  /// Statistics y_{jm} for state vectors v_{jm}. dimension is [J][M_{j}[S].
  std::vector< Matrix<double> > y_;
  /// Gaussian occupancies gamma_{jmi} for each substate. Dim is [J][M_{j}][I].
  std::vector< Matrix<double> > gamma_;
  
  double total_frames_, total_like_;

  /// Dimensionality of various subspaces
  int32 feature_dim_, phn_dim_, chn_dim_;
  int32 num_gauss_, num_states_;  ///< Other model specifications

  BaseFloat rand_prune_;
  KALDI_DISALLOW_COPY_AND_ASSIGN(MleAmPldaAccs);
  friend class MleAmPldaUpdater;
  friend class EbwAmPldaUpdater;
};

/** \class MleAmPldaUpdater
 *  Contains the functions needed to update the AmPlda parameters.
 */
class MleAmPldaUpdater {
 public:
  explicit MleAmPldaUpdater(const MleAmPldaOptions &options)
      : update_options_(options) {}
  void Reconfigure(const MleAmPldaOptions &options) {
    update_options_ = options;
  }

  /// Main update function: Computes some overall stats, does parameter updates
  /// and returns the total improvement of the different auxiliary functions.
  double Update(const MleAmPldaAccs &accs,
                   AmPlda *model,
                   PldaUpdateFlagsType flags);

  ///  Compute the Q_i quantities (Eq. 64).
  static void ComputeQ(const MleAmPldaAccs &accs,
                       const AmPlda &model,
                       std::vector< SpMatrix<double> > *Q);
		       
 private:
  MleAmPldaOptions update_options_;
  /// Q_{i}, quadratic term for phonetic subspace estimation. Dim is [I][S][S]
  std::vector< SpMatrix<double> > Q_;
  
  Vector<double> gamma_j_;  ///< State occupancies

   ///update projection matrices U
   double UpdateU(const MleAmPldaAccs &accs, AmPlda *model);
    
   ///update projection matrice G
   double UpdateG(const MleAmPldaAccs &accs, AmPlda *model);
      
   ///update bias vectors b
   double UpdateBias(const MleAmPldaAccs &accs, AmPlda  *model);
      
   ///update the residual variance vector
   double UpdateVars(const MleAmPldaAccs &accs, AmPlda  *model);
      
   ///update weighs 
   double UpdateWeights(const MleAmPldaAccs &accs, AmPlda  *model);
      
   double UpdatePhoneVectors(const MleAmPldaAccs &accs,
                                                   AmPlda *model,
                                                   const std::vector< SpMatrix<double> > &H);
						   
   void UpdateChannelPrior(const MleAmPldaAccs &accs, AmPlda *model);
   
   void RenormalizeV(const MleAmPldaAccs &accs,
                                    AmPlda *model,
                                    const SpMatrix<double> &H_sm);
				 
   void ComputeSmoothingTerms(const MleAmPldaAccs &accs,
                                             const AmPlda &model,
                                             const vector<SpMatrix<double> > &H,
                                             SpMatrix<double> *H_sm) const;
                                    
  KALDI_DISALLOW_COPY_AND_ASSIGN(MleAmPldaUpdater);
  MleAmPldaUpdater() {}  // Prevent unconfigured updater.
};


}  // End namespace kaldi


#endif  // KALDI_PLAD_MLE_AM_PLDA_H_
