
#include "plda/fa-gmm.h"
#include "plda/mle-fa-gmm.h"

namespace kaldi {
 
 FaGmmFlagsType StringToFaGmmFlags(std::string str) {
  FaGmmFlagsType flags = 0;
  for (const char *c = str.c_str(); *c != '\0'; c++) {
    switch (*c) {
      case 'U': flags |= kFaGmmProjection; break;
      case 'v': flags |= kFaGmmVars; break;
      case 'w': flags |= kFaGmmWeights; break;
      case 'b': flags |= kFaGmmBias; break;
      case 'G': flags |= kFaGmmStateProjection; break;
      default: KALDI_ERR << "Invalid element " << CharToString(*c)
                         << " of FaGmmFlagsType option string "
                         << str;
    }
  }
  return flags;
}


 void AccFaGmm::Write(std::ostream &out_stream, bool binary) const {
  uint32 tmp_uint32;
  
  WriteToken(out_stream, binary, "<FaGmmAcc>");
  WriteToken(out_stream, binary, "<NUMGaussians>");
  tmp_uint32 = static_cast<uint32>(num_gauss_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<FeatureDim>");
  tmp_uint32 = static_cast<uint32>(feature_dim_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<SpeechDim>");
  tmp_uint32 = static_cast<uint32>(speech_dim_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<StateDim>");
  tmp_uint32 = static_cast<uint32>(state_dim_);
  WriteBasicType(out_stream, binary, tmp_uint32);
  WriteToken(out_stream, binary, "<FLAGS>");
  WriteBasicType(out_stream, binary, flags_);
  
  if (!binary) out_stream << "\n";

  WriteToken(out_stream, binary, "<occ>");
  occupancy_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<b_acc>");
  b_accumulator_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<var_acc>");
  var_accumulator_.Write(out_stream, binary);
  WriteToken(out_stream, binary, "<first_order>");
  first_order_acc_.Write(out_stream, binary);
  
  if (second_order_acc_.size() != 0) {
    WriteToken(out_stream, binary, "<second_order>");
    for (int32 i = 0; i < num_gauss_; i++) {
      second_order_acc_[i].Write(out_stream, binary);
    }
  }
  
  if (U_accumulator_first_.size() != 0) {
    WriteToken(out_stream, binary, "<U_first>");
    for (int32 i = 0; i < num_gauss_; i++) {
      U_accumulator_first_[i].Write(out_stream, binary);
    }
  }
  
  if (U_accumulator_second_.size() != 0) {
    WriteToken(out_stream, binary, "<U_second>");
    for (int32 i = 0; i < num_gauss_; i++) {
      U_accumulator_second_[i].Write(out_stream, binary);
    }
  }
  
  if (G_accumulator_first_.size() != 0) {
    WriteToken(out_stream, binary, "<G_first>");
    for (int32 i = 0; i < num_gauss_; i++) {
      G_accumulator_first_[i].Write(out_stream, binary);
    }
  }
  
  if (G_accumulator_second_.size() != 0) {
    WriteToken(out_stream, binary, "<G_second>");
    for (int32 i = 0; i < num_gauss_; i++) {
      G_accumulator_second_[i].Write(out_stream, binary);
    }
  }

  WriteToken(out_stream, binary, "</FaGmmAcc>");
 }
 
 void AccFaGmm::Read(std::istream &in_stream, bool binary, bool add){
   uint32 tmp_uint32;
   std::string token;
   FaGmmFlagsType flags;
   int32 num_gauss, speech_dim, state_dim, feature_dim;
   
   ExpectToken(in_stream, binary, "<FaGmmAcc>");
   ExpectToken(in_stream, binary, "<NUMGaussians>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   num_gauss = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<FeatureDim>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   feature_dim = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<SpeechDim>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   speech_dim = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<StateDim>");
   ReadBasicType(in_stream, binary, &tmp_uint32);
   state_dim = static_cast<int32>(tmp_uint32);
   ExpectToken(in_stream, binary, "<FLAGS>");
   ReadBasicType(in_stream, binary, &flags);
  
   if (add) {
    if ((NumGauss() != 0 || FeatureDim() != 0 || SpeechDim() != 0)) {
      if (num_gauss != NumGauss() || speech_dim != SpeechDim() || state_dim != StateDim()
	|| feature_dim != FeatureDim() || flags != Flags())
        KALDI_ERR << "AccFaGmm::Read, dimension or flags mismatch, "
                  << NumGauss() << ", " << SpeechDim() << ", "
                  << FeatureDim() << " vs. " << num_gauss << ", "
                  << speech_dim << ", " << feature_dim << " (mixing accs from different "
                  << "models?";
    } else {
      Resize(num_gauss, feature_dim, speech_dim, state_dim, flags);
    }
  } else {
    Resize(num_gauss, feature_dim, speech_dim, state_dim, flags);
  }
  
   ExpectToken(in_stream, binary, "<occ>");
   occupancy_.Read(in_stream, binary, add);
   ExpectToken(in_stream, binary, "<b_acc>");
   b_accumulator_.Read(in_stream, binary, add);
   ExpectToken(in_stream, binary, "<var_acc>");
   var_accumulator_.Read(in_stream, binary, add);
   ExpectToken(in_stream, binary, "<first_order>");
   first_order_acc_.Read(in_stream, binary, add);
   
   ReadToken(in_stream, binary, &token);
   while (token != "</FaGmmAcc>") {
      if (token == "<U_first>") {
        for (size_t i = 0; i < U_accumulator_first_.size(); i++) {
           U_accumulator_first_[i].Read(in_stream, binary, add);
        }
      } else if (token == "<U_second>") {
        for (size_t i = 0; i < U_accumulator_second_.size(); i++) {
          U_accumulator_second_[i].Read(in_stream, binary, add);
        }
      } else if (token == "<second_order>") {
	for (size_t i = 0; i < second_order_acc_.size(); i++) {
	  second_order_acc_[i].Read(in_stream, binary, add);
	}
      } else if (token == "<G_first>") {
	 for (size_t i = 0; i < G_accumulator_first_.size(); i++) {
           G_accumulator_first_[i].Read(in_stream, binary, add);
        }
      } else if (token == "<G_second>") {
	for (size_t i = 0; i < G_accumulator_second_.size(); i++) {
          G_accumulator_second_[i].Read(in_stream, binary, add);
        }
      }
      
      ReadToken(in_stream, binary, &token);
   }
 }
 
 
 void AccFaGmm::Resize(int32 num_gauss, int32 feature_dim, 
		  int32 speech_dim, int32 state_dim, FaGmmFlagsType flags) {
   KALDI_ASSERT(num_gauss > 0 && 
		    feature_dim > 0 && 
		    speech_dim > 0 && 
		    state_dim >0);
		    
   num_gauss_ = num_gauss;
   feature_dim_ = feature_dim;
   speech_dim_ = speech_dim;
   state_dim_ = state_dim;
   flags_ = flags;
   
   occupancy_.Resize(num_gauss);
   first_order_acc_.Resize(num_gauss, feature_dim);
   second_order_acc_.resize(num_gauss);
   for(int32 m=0; m<num_gauss; m++) {
     second_order_acc_[m].Resize(feature_dim);  
   }
   
   if (flags_ & kFaGmmProjection) {
     U_accumulator_first_.resize(num_gauss);
     U_accumulator_second_.resize(num_gauss);
     for(int32 m=0; m<num_gauss; m++) {
       U_accumulator_first_[m].Resize(feature_dim, speech_dim);
       U_accumulator_second_[m].Resize(speech_dim);
     }
   } else {
    U_accumulator_first_.resize(0);
    U_accumulator_second_.resize(0);
   }
   
    if (flags_ & kFaGmmStateProjection) {
     G_accumulator_first_.resize(num_gauss);
     G_accumulator_second_.resize(num_gauss);
     for(int32 m=0; m<num_gauss; m++) {
       G_accumulator_first_[m].Resize(feature_dim, state_dim);
       G_accumulator_second_[m].Resize(state_dim);
     }
   } else {
    G_accumulator_first_.resize(0);
    G_accumulator_second_.resize(0);
   }
   
   if (flags_ & kFaGmmBias)
     b_accumulator_.Resize(num_gauss, feature_dim);
   else
     b_accumulator_.Resize(0, 0);
  
    if (flags_ & kFaGmmVars)
      var_accumulator_.Resize(num_gauss, feature_dim);
    else 
      var_accumulator_.Resize(0, 0);
 }
 
 void AccFaGmm::Resize(const FaGmm &model, FaGmmFlagsType flags) {
   Resize(model.NumGauss(), model.FeatureDim(), model.SpeechDim(), model.StateDim(), flags);
 }
 
void AccFaGmm::AccumulateForComponent(Vector<double> const &data,
						      const FaGmm &model,
						      int32 m,
						      double posterior,
						      const Vector<double> &x_mean,
						      const SpMatrix<double> &x_var) {

  int32 feature_dim = model.FeatureDim();
  
  KALDI_ASSERT(data.Dim() == feature_dim);
  KALDI_ASSERT(m < model.NumGauss());
  
  double wt = posterior;
  Vector<double> b(model.BiasVector(m));
  Matrix<double> U(model.Projection(m));
  Vector<double> U_xt(feature_dim);
 // SpMatrix<double> x_var(model.InverseV(m));
  
  // accumulate
  occupancy_(m) += wt;
  first_order_acc_.Row(m).AddVec(wt, data);
  second_order_acc_[m].AddVec2(wt, data);
  
  Vector<double> data_bias(data);  
  data_bias.AddVec(-1, b);   //y_t - b_m
  U_xt.AddMatVec(1.0, U, kNoTrans, x_mean, 0.0);
  
  if (flags_ & kFaGmmProjection) {
    U_accumulator_first_[m].AddVecVec(wt, data_bias, x_mean);
    U_accumulator_second_[m].AddVec2(wt, x_mean);
    if(mConfig_.usePostVar) {
      U_accumulator_second_[m].AddSp(wt, x_var);
    }
  }
  
  if (flags_ & kFaGmmBias) {
    b_accumulator_.Row(m).AddVec(wt, data);
    b_accumulator_.Row(m).AddVec(-wt, U_xt);
  }  
   
   if(flags_ & kFaGmmVars) {
     data_bias.AddVec(-1.0, U_xt);
     var_accumulator_.Row(m).AddVecVec(wt, data_bias, data_bias, 1.0);
     if(mConfig_.usePostVar) {
//	SpMatrix<double> tmp_mat(feature_dim);
//	tmp_mat.AddMat2Sp(1.0, U, kNoTrans, x_var, 0.0);
//	Vector<double> diag(feature_dim);
//	diag.CopyDiagFromSp(tmp_mat);
	var_accumulator_.Row(m).AddVec(wt, model.VarOffset().Row(m));
      }
    }
  
}

void AccFaGmm::AccumulateForComponentPLDA(Vector<double> const &data,
						      const FaGmm &model,
						      int32 m,
						      double posterior,
						      const Vector<double> &x_mean,
//						      const SpMatrix<double> &x_var,
						      const Vector<double> &z_mean,
						      const SpMatrix<double> &z_var) {
  
  int32 feature_dim = model.FeatureDim();
  
  KALDI_ASSERT(data.Dim() == feature_dim);
  KALDI_ASSERT(m < model.NumGauss());
  
  double wt = posterior;
  Vector<double> b(model.BiasVector(m));
  Matrix<double> U(model.Projection(m));
  Matrix<double> G(model.StateProjection(m));
  Vector<double> U_xt(feature_dim);
  Vector<double> G_zt(feature_dim);
  SpMatrix<double> x_var(model.InverseV(m));
  
  // accumulate
  occupancy_(m) += wt;
  first_order_acc_.Row(m).AddVec(wt, data);
  second_order_acc_[m].AddVec2(wt, data);
  
  Vector<double> data_bias(data);  
  data_bias.AddVec(-1, b);   //y_t - b_m
  U_xt.AddMatVec(1.0, U, kNoTrans, x_mean, 0.0);
  G_zt.AddMatVec(1.0, G, kNoTrans, z_mean, 0.0);
  
  if (flags_ & kFaGmmProjection) {
    Vector<double> tmp_vec(data_bias);
    tmp_vec.AddVec(-1.0, G_zt);
    U_accumulator_first_[m].AddVecVec(wt, tmp_vec, x_mean);
    U_accumulator_second_[m].AddVec2(wt, x_mean);
    if(mConfig_.usePostVar) {
      U_accumulator_second_[m].AddSp(wt, x_var);
    }
  }
  
  if (flags_ & kFaGmmStateProjection) {
    Vector<double> tmp_vec(data_bias);
    tmp_vec.AddVec(-1.0, U_xt);
    G_accumulator_first_[m].AddVecVec(wt, tmp_vec, z_mean);
    G_accumulator_second_[m].AddVec2(wt, z_mean);
 //   KALDI_LOG<<z_mean;
//    KALDI_LOG<<G_accumulator_first_[m]<<G_accumulator_second_[m];
    if(mConfig_.usePostVar) {
      G_accumulator_second_[m].AddSp(wt, z_var);
    }
  }
  
  if (flags_ & kFaGmmBias) {
    b_accumulator_.Row(m).AddVec(wt, data);
    b_accumulator_.Row(m).AddVec(-wt, U_xt);
    b_accumulator_.Row(m).AddVec(-wt, G_zt);
  }  
   
   if(flags_ & kFaGmmVars) {
     data_bias.AddVec(-1.0, U_xt);
     data_bias.AddVec(-1.0, G_zt);
     var_accumulator_.Row(m).AddVecVec(wt, data_bias, data_bias, 1.0);
     if(mConfig_.usePostVar) {
	SpMatrix<double> tmp_mat(feature_dim);
	tmp_mat.AddMat2Sp(1.0, U, kNoTrans, x_var, 0.0);
	if(flags_ & kFaGmmStateProjection)
	  tmp_mat.AddMat2Sp(1.0, G, kNoTrans, z_var, 1.0);
	Vector<double> diag(feature_dim);
	diag.CopyDiagFromSp(tmp_mat);
	var_accumulator_.Row(m).AddVec(wt, diag);
      }
    }
    
}

//NOTE: call this function after compute the inverse V by FaGMM
//and updateSigma()
double AccFaGmm::AccumulatePLDA(Vector<double> const &data,
				 FaGmm &model,
				 const std::vector<int32> &gselect,
			         const Vector<double> &z_mean,
				 const SpMatrix<double> &z_var) {
  
  Vector<double> posterior;
  double loglike;
  
  loglike = model.ComponentPosteriors(data, gselect, &posterior);
//  KALDI_LOG<<"posterior = "<<posterior;
  
  for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];
      double this_posterior = posterior(ki);
      if(this_posterior > posterior_threshold) {
        const Vector<double> x_mean = model.PosteriorFrameMeanPLDA(data, z_mean, m);
 //     KALDI_LOG<<"m = "<<m<<" xt = "<<x_mean;
        AccumulateForComponentPLDA(data, model, m, this_posterior, 
						    x_mean, z_mean, z_var);
      }
    }
    
  return loglike;
}

//NOTE: call this function after compute the inverse V by FaGMM
double AccFaGmm::Accumulate(Vector<double> const &data,
				 FaGmm &model,
				 const std::vector<int32> &gselect) {
  
  Vector<double> posterior;
  double loglike;
  
  loglike = model.ComponentPosteriors(data, gselect, &posterior);
//  KALDI_LOG<<"posterior = "<<posterior;
  
  for(int32 ki = 0, last = gselect.size();  ki < last; ki++ ) {
      int32 m = gselect[ki];
      double this_posterior = posterior(ki);
      if(this_posterior > posterior_threshold) {
        const Vector<double> x_mean = model.CondPosteriorMean(data, m);
 //     KALDI_LOG<<"m = "<<m<<" xt = "<<x_mean;
        AccumulateForComponent(data, model, m, this_posterior, x_mean, model.InverseV(m));
      }
    }
    
  return loglike;
}

///returen the average loglikelihood per frame
double MleFaGmmUpdater::MlObjective(const AccFaGmm &fa_gmm_acc, 
				       FaGmm &model) const {
 
  double acc_sum = fa_gmm_acc.occupancy().Sum();  //we normalize the likelihood in case of overflow
  Vector<double> occ_bf(fa_gmm_acc.occupancy());
  double obj = VecVec(occ_bf, model.gconsts()) / acc_sum;
  
  for(int32 m=0; m<model.NumGauss(); m++) {
    Vector<double> bias = model.BiasVector(m);
    SpMatrix<double> inv_Sigma = model.Inv_Sigma(m);
    
    obj += VecSpVec(bias, inv_Sigma, fa_gmm_acc.first_order_acc().Row(m)) / acc_sum;
    obj -= 0.5 * occ_bf(m) * VecSpVec(bias, inv_Sigma, bias) / acc_sum;
    obj -= 0.5 * TraceSpSp(inv_Sigma, fa_gmm_acc.second_order_acc(m)) / acc_sum;
  }
  
  return obj;
}

double MleFaGmmUpdater::UpdateU(const AccFaGmm &fa_gmm_acc, FaGmm *model) {
  
  int32 num_gauss = model->NumGauss();
  int32 feature_dim = model->FeatureDim();
  int32 speech_dim = model->SpeechDim();
  Matrix<double> U_m(feature_dim, speech_dim);
  SpMatrix<double> tmp_mat(speech_dim);
  SpMatrix<double> this_var(feature_dim);
  double obj_old, obj_new;
  
  if(mConfig_.checkObj) {
    //get the objective value before updating U
    obj_old=MlObjective(fa_gmm_acc, *model);
  }
  
  SolverOptions opts;
  opts.name = "U";
  opts.K = mConfig_.max_cond;
  opts.eps = mConfig_.epsilon;
    
  for(int m = 0; m<num_gauss; m++) {
    tmp_mat.CopyFromSp(fa_gmm_acc.U_accumulator_second(m));
  //  tmp_mat.InvertDouble();
  //  U_m.AddMatSp(1.0, fa_gmm_acc.U_accumulator_first(m), kNoTrans, tmp_mat, 0.0);
    this_var.SetZero();
    this_var.AddVec(1.0, model->InvCovars(m));
    double impr =
        SolveQuadraticMatrixProblem(tmp_mat, fa_gmm_acc.U_accumulator_first(m),
                                    this_var, opts, &U_m);
				    
  KALDI_LOG<<"Trace UU^T before update "<<TraceMatMat(model->U_[m], model->U_[m], kTrans);
    model->U_[m].CopyFromMat(U_m);
    
  KALDI_LOG<<"Trace UU^T after update "<<TraceMatMat(model->U_[m], model->U_[m], kTrans);
  //  if (m < 10) {
      KALDI_VLOG(2) << "Objf impr for projection U for m = " << m << ", is "
                    << (impr/(fa_gmm_acc.occupancy()(m) + 1.0e-20)) << " over " << fa_gmm_acc.occupancy()(m)
                    << " frames";
//    }
  }
  
   model->ComputeSigma();
   if(mConfig_.checkObj) {
     //get the objective value after updating U
    obj_new = MlObjective(fa_gmm_acc, *model);
    double obj_impr = obj_new - obj_old;
    KALDI_LOG<<"average loglike improvement for updating U is "<< obj_impr;
    //return the avg improvement per frame 
    return obj_impr;
   } else {
     return 0.0;
   }
}

double MleFaGmmUpdater::UpdateG(const AccFaGmm &fa_gmm_acc, FaGmm *model) {
  
  int32 num_gauss = model->NumGauss();
  int32 feature_dim = model->FeatureDim();
  int32 state_dim = model->StateDim();
  Matrix<double> G_m(feature_dim, state_dim);
  SpMatrix<double> tmp_mat(state_dim);
  SpMatrix<double> this_var(feature_dim);
  double obj_old, obj_new;
  
  if(mConfig_.checkObj) {
    //get the objective value before updating G
    obj_old=MlObjective(fa_gmm_acc, *model);
  }
  
   if (model->G_.size() == 0) model->G_.resize(num_gauss);
   
  SolverOptions opts;
  opts.name = "G";
  opts.K = mConfig_.max_cond;
  opts.eps = mConfig_.epsilon;
  
  for(int m = 0; m<num_gauss; m++) {
    tmp_mat.CopyFromSp(fa_gmm_acc.G_accumulator_second(m));
  //  tmp_mat.InvertDouble();
  //  U_m.AddMatSp(1.0, fa_gmm_acc.U_accumulator_first(m), kNoTrans, tmp_mat, 0.0);
  //  KALDI_LOG<<fa_gmm_acc.G_accumulator_first(m);
    this_var.SetZero();
    this_var.AddVec(1.0, model->InvCovars(m));
    double impr =
        SolveQuadraticMatrixProblem(tmp_mat, fa_gmm_acc.G_accumulator_first(m),
                                    this_var, opts, &G_m);
				    
  KALDI_LOG<<"Trace GG^T before update "<<TraceMatMat(model->G_[m], model->G_[m], kTrans);
  model->G_[m].Resize(feature_dim, state_dim);
  model->G_[m].CopyFromMat(G_m);
    
  KALDI_LOG<<"Trace GG^T after update "<<TraceMatMat(model->G_[m], model->G_[m], kTrans);
  //  if (m < 10) {
      KALDI_VLOG(2) << "Objf impr for projection G for m = " << m << ", is "
                    << (impr/(fa_gmm_acc.occupancy()(m) + 1.0e-20)) << " over " << fa_gmm_acc.occupancy()(m)
                    << " frames";
//    }
  }
  
//   model->UpdateSigmaPLDA();
   if(mConfig_.checkObj) {
     //get the objective value after updating U
    obj_new = MlObjective(fa_gmm_acc, *model);
    double obj_impr = obj_new - obj_old;
    KALDI_LOG<<"average loglike improvement for updating G is "<< obj_impr;
    //return the avg improvement per frame 
    return obj_impr;
   } else {
     return 0.0;
   }
}

double MleFaGmmUpdater::UpdateBias(const AccFaGmm &fa_gmm_acc, FaGmm *model) {
  
  int32 num_gauss = model->NumGauss();
  double obj_old, obj_new;
  
  if(mConfig_.checkObj) {
    //get the objective value before updating b
    obj_old = MlObjective(fa_gmm_acc, *model);
  }
  
  for(int m=0; m<num_gauss; m++) {
    model->b_.Row(m).CopyFromVec(fa_gmm_acc.b_accumulator().Row(m));
    model->b_.Row(m).Scale(1.0/fa_gmm_acc.occupancy()(m));
  }
 
//   model->ComputeGconsts(); 
   if(mConfig_.checkObj) {
    //get the objective value after updating b
    obj_new = MlObjective(fa_gmm_acc, *model);
    double obj_impr = obj_new - obj_old;
    KALDI_LOG<<"average loglike improvement for updating b is "<< obj_impr;
    //return the avg improvement per frame 
    return obj_impr;
   } else {
     return 0.0;
   }
  
}

double MleFaGmmUpdater::UpdateVars(const AccFaGmm &fa_gmm_acc, FaGmm *model) {
  
  int32 num_gauss = model->NumGauss();
  double obj_old, obj_new;
  int32 tot_floored = 0, gauss_floored = 0;
  
  if(mConfig_.checkObj) {
    //get the objective value before updating \lambda
    obj_old = MlObjective(fa_gmm_acc, *model);
  }
  
  for(int m=0; m<num_gauss; m++) {
    model->inv_vars_.Row(m).CopyFromVec(fa_gmm_acc.var_accumulator().Row(m));
    model->inv_vars_.Row(m).Scale(1.0/fa_gmm_acc.occupancy()(m));
    
    int32 floored;
    if (mConfig_.variance_floor_vector.Dim() != 0) {
       floored = model->inv_vars_.Row(m).ApplyFloor(mConfig_.variance_floor_vector);
    } else {
       floored = model->inv_vars_.Row(m).ApplyFloor(mConfig_.min_variance);
    }
     
     if (floored != 0) {
       tot_floored += floored;
       gauss_floored++;
     }
    model->inv_vars_.Row(m).InvertElements();
  }

  if (tot_floored > 0)
    KALDI_WARN << tot_floored << " variances floored in " << gauss_floored
               << " Gaussians.";
	       
  model->ComputeSigma(); 
   if(mConfig_.checkObj) {
    //get the objective value after updating \lambda
    obj_new = MlObjective(fa_gmm_acc, *model);
    double obj_impr = obj_new - obj_old;
    KALDI_LOG<<"average loglike improvement for updating convariance is "<< obj_impr;
    //return the avg improvement per frame 
    return obj_impr;
   } else {
     return 0.0;
   }
   
}

double MleFaGmmUpdater::UpdateWeights(const AccFaGmm &fa_gmm_acc, FaGmm *model){
  
  int32 num_gauss = model->NumGauss();
  double obj_old, obj_new;
  double occ_sum = fa_gmm_acc.occupancy().Sum();
  
  if(mConfig_.checkObj) {
    //get the objective value before updating w
    obj_old = MlObjective(fa_gmm_acc, *model);
  }
  
  for (int32 m = 0; m < num_gauss; m++) {
    double occ = fa_gmm_acc.occupancy()(m);
    double prob;
    if (occ_sum > 0.0)
      prob = occ / occ_sum;
    else
      prob = 1.0 / num_gauss;
      
      if (occ > mConfig_.min_gaussian_occupancy  
	&& prob > mConfig_.min_gaussian_weight)
      model->weights_(m) = prob;
      else {
	model->weights_(m) = mConfig_.min_gaussian_weight;
        KALDI_WARN << "Gaussian has too little data m = "
                  << m << ", occ = " << occ << ", weight = " << prob;	   
       }
  }
  
    model->ComputeGconsts(); 
    if(mConfig_.checkObj) {
    //get the objective value after updating w
    obj_new = MlObjective(fa_gmm_acc, *model);
    double obj_impr = obj_new - obj_old;
    KALDI_LOG<<"average loglike improvement for updating w is "<< obj_impr;
    //return the avg improvement per frame 
    return obj_impr;
   } else {
     return 0.0;
   }
   
  }


double MleFaGmmUpdater::Update(const AccFaGmm &fa_gmm_acc, FaGmm *model) {
                      
  KALDI_ASSERT(model != NULL);

  double loglike_improve = 0.0;
  
  if (flags_ & ~fa_gmm_acc.Flags())
    KALDI_ERR << "Flags in argument do not match the active accumulators";

  KALDI_ASSERT(fa_gmm_acc.NumGauss() == model->NumGauss() &&
               fa_gmm_acc.FeatureDim() == model->FeatureDim());

 // KALDI_LOG<<fa_gmm_acc.occupancy();
  
  if (flags_ &  kFaGmmProjection) {
     loglike_improve += UpdateU(fa_gmm_acc, model);
  }
  
  if (flags_ & kFaGmmBias) {
    loglike_improve += UpdateBias(fa_gmm_acc, model);
  }
       
  if (flags_ & kFaGmmVars) {
    loglike_improve += UpdateVars(fa_gmm_acc, model);
  }
  
  if (flags_ & kFaGmmWeights) {
    loglike_improve += UpdateWeights(fa_gmm_acc, model);
  }
 
  if (flags_ &  kFaGmmStateProjection) {
     loglike_improve += UpdateG(fa_gmm_acc, model);
  }
  
  if(flags_ & kFaGmmProjection || flags_ & kFaGmmVars) {
     model->ComputeV();  //update the matrix inv_V
     model->ComputeH(); //update H matrices
  }
   
  return loglike_improve;
}

}//end of namespace
