


#ifndef KALDI_MLE_FA_GMM_H_
#define KALDI_MLE_FA_GMM_H_ 1

#include "itf/options-itf.h"
#include "plda/fa-gmm.h"
#include "matrix/matrix-lib.h"
#include "base/kaldi-common.h"

namespace kaldi {
  
  enum FaGmmUpdateFlags {
    kFaGmmProjection      = 0x001,  // U
    kFaGmmVars 		= 0x002,  // v
    kFaGmmWeights     	= 0x004,  // w
    kFaGmmBias		= 0x008,  // b
    kFaGmmStateProjection             = 0x010,  // G
  };
  
typedef uint16 FaGmmFlagsType;

FaGmmFlagsType StringToFaGmmFlags(std::string str);

/** \struct MleFaGmmOptions
 *  Configuration variables like variance floor, minimum occupancy, etc.
 *  needed in the estimation process.
 */
struct MleFaGmmOptions {
  /// Variance floor for each dimension [empty if not supplied].
  /// It is in double since the variance is computed in double precision.
  Vector<double> variance_floor_vector;
  /// Minimum weight below which a Gaussian is not updated (and is
  /// removed, if remove_low_count_gaussians == true);
  double min_gaussian_weight;
  /// Minimum count below which a Gaussian is not updated (and is
  /// removed, if remove_low_count_gaussians == true).
  double min_gaussian_occupancy;
  /// Minimum allowed variance in any dimension (if no variance floor)
  /// It is in double since the variance is computed in double precision.
  double min_variance;
//  double posterior_threshold;  ///accumulate stats when posterior is above the threshold
  bool usePostVar;    ///use posterior variance of p(x|y,m) to accumulate stats
  bool checkObj; 	///check the objetive function when updating the model paramters
  
  double max_cond;
  double epsilon;
 // bool remove_low_count_gaussians;
  MleFaGmmOptions() {
    // don't set var floor vector by default.
    min_gaussian_weight     = 1.0e-05;
    min_gaussian_occupancy  = 10.0;
    min_variance            = 0.0001;
    usePostVar 	= true;
    checkObj	= true;
    max_cond 	= 1.0e+05;
    epsilon 	= 1.0e-40;
  }
  void Register(OptionsItf *po) {
    std::string module = "MleDiagGmmOptions: ";
    po->Register("min-gaussian-weight", &min_gaussian_weight,
                 module+"Min Gaussian weight before we remove it.");
    po->Register("min-gaussian-occupancy", &min_gaussian_occupancy,
                 module+"Minimum occupancy to update a Gaussian.");
    po->Register("min-variance", &min_variance,
                 module+"Variance floor (absolute variance).");
    po->Register("use-post-var", &usePostVar,
                 module+"If true, use posterior variance p(x|y,m) to accumulate stats.");
    po->Register("check-objetive-func", &checkObj,
                 module+"If true, compute the objective function before and after update the model parameters.");
  }
};

  class AccFaGmm {
    
    public:
      
      AccFaGmm(): num_gauss_(0), feature_dim_(0), speech_dim_(0), flags_(0) { }
      
      explicit AccFaGmm(const FaGmm &fagmm, FaGmmFlagsType flags) {
          Resize(fagmm, flags);
      }
     
      ///IO functions
      void Read(std::istream &in_stream, bool binary, bool add);
      void Write(std::ostream &out_stream, bool binary) const;
      
      ///resize the accumulator
      void Resize(int32 num_gauss, int32 feature_dim, 
		  int32 speech_dim, int32 state_dim, FaGmmFlagsType flags);
      /// resize the accumulator with arguments based on fagmm
      void Resize(const FaGmm &model, FaGmmFlagsType flags);
  
      void ReConfig(MleFaGmmOptions config) { mConfig_ = config; }
      
      ///accumulate a single component 'm', given the poisterior and 
      ///mean and variance of p(x|y,m)
      void AccumulateForComponent(Vector<double> const &data,
						      const FaGmm &model,
						      int32 m,
						      double posterior,
						      const Vector<double> &x_mean,
						      const SpMatrix<double> &x_var);
						      
      void AccumulateForComponentPLDA(Vector<double> const &data,
						      const FaGmm &model,
						      int32 m,
						      double posterior,
						      const Vector<double> &x_mean,
	//					      const SpMatrix<double> &x_var,
						      const Vector<double> &z_mean,
						      const SpMatrix<double> &z_var);
						      
      ///accumulate for all components				      
      double Accumulate(Vector<double> const &data,
				 FaGmm &gmm,
				 const std::vector<int32> &gselect);
      
     ///accumulate for all components for PLDA model
     double AccumulatePLDA(Vector<double> const &data,
				 FaGmm &model,
				 const std::vector<int32> &gselect,
			         const Vector<double> &z_mean,
				 const SpMatrix<double> &z_var);
				 
      ///Accessors
      Matrix<double> U_accumulator_first(int32 m) const { return U_accumulator_first_[m]; }
      SpMatrix<double> U_accumulator_second(int32 m) const { return U_accumulator_second_[m]; }
      Matrix<double> G_accumulator_first(int32 m) const { return G_accumulator_first_[m]; }
      SpMatrix<double> G_accumulator_second(int32 m) const { return G_accumulator_second_[m]; }
      
      Vector<double> occupancy() const { return occupancy_; }
      Matrix<double> b_accumulator() const { return b_accumulator_; }
      Matrix<double> var_accumulator() const { return var_accumulator_; }
      Matrix<double> first_order_acc() const {return first_order_acc_; }
      SpMatrix<double> second_order_acc(int32 m) const {return second_order_acc_[m]; } 
      
      int32 NumGauss() const { return num_gauss_; }
      int32 FeatureDim() const { return feature_dim_; }
      int32 SpeechDim() const { return speech_dim_; }
      int32 StateDim() const { return state_dim_; }
      FaGmmFlagsType Flags() const {return flags_; }
      
    private:
      
      int32 num_gauss_; 
      int32 feature_dim_;
      int32 speech_dim_;
      int32 state_dim_;
      
      FaGmmFlagsType flags_;
      MleFaGmmOptions mConfig_;
      
      Vector<double> occupancy_;
      ///First part of the stats to update U: sum_t w_{mt}(y_t - b_m)E^T[x_t]
      std::vector< Matrix<double> > U_accumulator_first_;  
      ///Second part of the stats to update U: sum_t w_{mt} E[x_t*x_t^T]
      std::vector< SpMatrix<double> > U_accumulator_second_;
      ///First part of the stats to update G
      std::vector< Matrix<double> > G_accumulator_first_;
      ///Second part of the stats to update G
      std::vector< SpMatrix<double> > G_accumulator_second_;
      ///Statistics to update bias vector b_m
      Matrix<double> b_accumulator_;
      ///Statistics to update the variance \Lambda_m
      Matrix<double> var_accumulator_;
      ///First order statistics w_{mt}y_t
      Matrix<double> first_order_acc_;
      ///Second order statistics w_{mt}y_t*y_t^T
      std::vector< SpMatrix<double> > second_order_acc_;
      
  };
  
  class MleFaGmmUpdater{
    
    public:
      ///empety constructor
      MleFaGmmUpdater() : flags_(0){}
      MleFaGmmUpdater(MleFaGmmOptions config, FaGmmFlagsType flags) {
	mConfig_ = config; flags_ = flags;
      }
      
      ///update the FaGmm model parameters by maximum likelihood
      double Update(const AccFaGmm &fa_gmm_acc, FaGmm *model);
  
      void ReConfig(MleFaGmmOptions config, FaGmmFlagsType flags) { mConfig_ = config; flags_ = flags; }
  
    private:
      
      FaGmmFlagsType flags_;
      MleFaGmmOptions mConfig_;
      
      ///compute the ML objective function to monitor the model update
      double MlObjective(const AccFaGmm &fa_gmm_acc, FaGmm &model) const;
      
      ///update projection matrices U
      double UpdateU(const AccFaGmm &fa_gmm_acc, FaGmm *model);
    
      ///update projection matrice G
      double UpdateG(const AccFaGmm &fa_gmm_acc, FaGmm *model);
      
      ///update bias vectors b
      double UpdateBias(const AccFaGmm &fa_gmm_acc, FaGmm *model);
      
      ///update the residual variance vector
      double UpdateVars(const AccFaGmm &fa_gmm_acc, FaGmm *model);
      
      ///update weighs 
      double UpdateWeights(const AccFaGmm &fa_gmm_acc, FaGmm *model);
      
  };
  
}//end of namespace
#endif
