

#include "base/kaldi-common.h"
#include "util/common-utils.h"
#include "gmm/model-common.h"
#include "gmm/full-gmm.h"
#include "gmm/diag-gmm.h"
#include "gmm/mle-full-gmm.h"
#include "plda/fa-gmm.h"
#include "plda/mle-fa-gmm.h"

int main(int argc, char *argv[]) {
  
  try {
    
    using namespace kaldi;

    const char *usage =
        "Accumulate stats for training a Fa-GMM.\n"
        "Usage:  fa-gmm-acc-stats [options] <model-in> <feature-rspecifier> "
        "<stats-out>\n"
        "e.g.: fa-gmm-acc-stats 1.mdl 1.gs scp:train.scp 1.acc\n";

    ParseOptions po(usage);
    MleFaGmmOptions mle_opts;
    bool binary_write = true;
    std::string update_flags_str = "Uvwb";
    std::string weights_rspecifier;
    po.Register("binary", &binary_write, "Write output in binary mode");
    po.Register("update-flags", &update_flags_str, "Which Fa-Gmm parameters will be "
                "updated: subset of Uvwb.");
 //   po.Register("gselect", &gselect_rspecifier, "rspecifier for gselect objects "
 //               "to limit the #Gaussians accessed on each frame.");
    po.Register("weights", &weights_rspecifier, "rspecifier for a vector of floats "
                "for each utterance, that's a per-frame weight.");
    mle_opts.Register(&po);
    po.Read(argc, argv);

    if (po.NumArgs() != 4) {
      po.PrintUsage();
      exit(1);
    }

    std::string model_filename = po.GetArg(1),
        gselect_rspecifier = po.GetArg(2),
        feature_rspecifier = po.GetArg(3),
        accs_wxfilename = po.GetArg(4);

    FaGmm fagmm;
    {
      bool binary_read;
      Input ki(model_filename, &binary_read);
      fagmm.Read(ki.Stream(), binary_read);
    }

    AccFaGmm fagmm_accs;
    fagmm_accs.Resize(fagmm, StringToFaGmmFlags(update_flags_str));
    fagmm_accs.ReConfig(mle_opts);
    
    double tot_like = 0.0, tot_weight = 0.0;

    SequentialBaseFloatMatrixReader feature_reader(feature_rspecifier);
    RandomAccessInt32VectorVectorReader gselect_reader(gselect_rspecifier);
    RandomAccessBaseFloatVectorReader weights_reader(weights_rspecifier);
    int32 num_done = 0, num_err = 0;

    for (; !feature_reader.Done(); feature_reader.Next()) {
      std::string key = feature_reader.Key();
      const Matrix<BaseFloat> &mat = feature_reader.Value();
      int32 file_frames = mat.NumRows();
      BaseFloat file_like = 0.0,
          file_weight = 0.0; // total of weights of frames (will each be 1 unless
      // --weights option supplied.
      Vector<BaseFloat> weights;
      if (weights_rspecifier != "") { // We have per-frame weighting.
        if (!weights_reader.HasKey(key)) {
          KALDI_WARN << "No per-frame weights available for utterance " << key;
          num_err++;
          continue;
        }
        weights = weights_reader.Value(key);
        if (weights.Dim() != file_frames) {
          KALDI_WARN << "Weights for utterance " << key << " have wrong dim "
                     << weights.Dim() << " vs. " << file_frames;
          num_err++;
          continue;
        }
      }
      
      if (!gselect_reader.HasKey(key)) {
        KALDI_WARN << "No gselect information for utterance " << key;
        num_err++;
        continue;
      }
        const std::vector<std::vector<int32> > &gselect =
            gselect_reader.Value(key);
        if (gselect.size() != static_cast<size_t>(file_frames)) {
          KALDI_WARN << "gselect information for utterance " << key
                     << " has wrong size " << gselect.size() << " vs. "
                     << file_frames;
          num_err++;
          continue;
        }
        
        for (int32 i = 0; i < file_frames; i++) {
          BaseFloat weight = (weights.Dim() != 0) ? weights(i) : 1.0;
          if (weight == 0.0) continue;
          file_weight += weight;
          Vector<double> data(mat.NumCols());
	  data.CopyRowFromMat(mat, i);
          const std::vector<int32> &this_gselect = gselect[i];
          int32 gselect_size = this_gselect.size();
          KALDI_ASSERT(gselect_size > 0);
//          std::vector<int32> this_gselect(1);
//	  this_gselect[0] = 0;

	  //KALDI_LOG<<data;	  
	  double loglike = fagmm_accs.Accumulate(data, fagmm, this_gselect);
          file_like += weight * loglike;

        }
      KALDI_VLOG(2) << "File '" << key << "': Average likelihood = "
                    << (file_like/file_weight) << " over "
                    << file_weight <<" frames.";
      tot_like += file_like;
      tot_weight += file_weight;
      num_done++;
    }
    KALDI_LOG << "Done " << num_done << " files; "
              << num_err << " with errors.";
    KALDI_LOG << "Overall likelihood per "
              << "frame = " << (tot_like/tot_weight) << " over " << tot_weight
              << " (weighted) frames.";
    {
	Output ko(accs_wxfilename, binary_write);
	fagmm_accs.Write(ko.Stream(), binary_write); 
	KALDI_LOG << "Written accs to " << accs_wxfilename;
    }
    return (num_done != 0 ? 0 : 1);
  } catch(const std::exception &e) {
    std::cerr << e.what();
    return -1;
  }
}
