// pldabin/plda-acc-stats-ali.cc

// Liang Lu (2014)

// See ../../COPYING for clarification regarding multiple authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.


#include "base/kaldi-common.h"
#include "util/common-utils.h"
#include "plda/am-plda.h"
#include "hmm/transition-model.h"
#include "plda/mle-am-plda.h"

int main(int argc, char *argv[]) {
  try {
    using namespace kaldi;
    const char *usage =
        "Accumulate stats for AmPlda training.\n"
        "Usage: plda-acc-stats-ali [options] <model-in> <feature-rspecifier> "
        "<alignments-rspecifier> <stats-out>\n"
        "e.g.: plda-acc-stats-ali 1.mdl scp:train.scp ark:1.gs ark:1.ali 1.acc\n";

    ParseOptions po(usage);
    bool binary = true;
    std::string utt2spk_rspecifier;
    std::string update_flags_str = "zUGbwvt";
    MleAmPldaOptions plda_opts;
    BaseFloat rand_prune = 1.0e-05;
    int shift_frame = 0;
    po.Register("binary", &binary, "Write output in binary mode");
  //  po.Register("gselect", &gselect_rspecifier, "Precomputed Gaussian indices (rspecifier)");
    po.Register("utt2spk", &utt2spk_rspecifier, "rspecifier for utterance to speaker map");
    po.Register("shift-frame", &shift_frame, "the step to shift the frame to accumuates stats, for multi-frame plda acoustic model");
    po.Register("rand-prune", &rand_prune, "Randomized pruning threshold for posteriors");
    po.Register("update-flags", &update_flags_str, "Which PLDA parameters to update: subset of zUGbwvt.");
    plda_opts.Register(&po);
    po.Read(argc, argv);

    if (po.NumArgs() != 5) {
      po.PrintUsage();
      exit(1);
    }

    kaldi::PldaUpdateFlagsType acc_flags = StringToPldaUpdateFlags(update_flags_str);

    std::string model_filename = po.GetArg(1),
        feature_rspecifier = po.GetArg(2),
        gselect_rspecifier = po.GetArg(3),
        alignments_rspecifier = po.GetArg(4),
        accs_wxfilename = po.GetArg(5);

    using namespace kaldi;
    typedef kaldi::int32 int32;

    AmPlda am_plda;
    TransitionModel trans_model;
    {
      bool binary;
      Input ki(model_filename, &binary);
      trans_model.Read(ki.Stream(), binary);
      am_plda.Read(ki.Stream(), binary);
    }

    Vector<double> transition_accs;
    if (acc_flags & kaldi::kPldaTransitions)
      trans_model.InitStats(&transition_accs);
    MleAmPldaAccs plda_accs(rand_prune);
    plda_accs.ResizeAccumulators(am_plda, acc_flags);

    double tot_like = 0.0;
    kaldi::int64 tot_t = 0;

    SequentialBaseFloatMatrixReader feature_reader(feature_rspecifier);
    RandomAccessInt32VectorReader alignments_reader(alignments_rspecifier);

    RandomAccessInt32VectorVectorReader gselect_reader;
    if (!gselect_rspecifier.empty() && !gselect_reader.Open(gselect_rspecifier))
      KALDI_ERR << "Unable to open stream for gaussian-selection indices";

    kaldi::PldaPerFrameDerivedVars per_frame_vars;

    int32 num_done = 0, num_no_alignment = 0, num_other_error = 0;
    for (; !feature_reader.Done(); feature_reader.Next()) {
      std::string utt = feature_reader.Key();
      if (!alignments_reader.HasKey(utt)) {
        num_no_alignment++;
      } else {
        const Matrix<BaseFloat> &mat = feature_reader.Value();
        const std::vector<int32> &alignment = alignments_reader.Value(utt);

        bool have_gselect  = !gselect_rspecifier.empty()
            && gselect_reader.HasKey(utt)
            && gselect_reader.Value(utt).size() == mat.NumRows();
        if (!gselect_rspecifier.empty() && !have_gselect)
          KALDI_WARN << "No Gaussian-selection info available for utterance "
                     << utt << " (or wrong size)\n";
        std::vector<std::vector<int32> > empty_gselect;
        const std::vector<std::vector<int32> > *gselect =
            (have_gselect ? &gselect_reader.Value(utt) : &empty_gselect);

        if (alignment.size() != mat.NumRows()) {
          KALDI_WARN << "Alignments has wrong size "<< (alignment.size()) <<
              " vs. "<< (mat.NumRows());
          num_other_error++;
          continue;
        }

        num_done++;
        BaseFloat tot_like_this_file = 0.0;

        for (size_t i = 0; i < alignment.size(); i++) {
          int32 tid = alignment[i],  // transition identifier.
              pdf_id = trans_model.TransitionIdToPdf(tid);
          if (acc_flags & kaldi::kPldaTransitions)
            trans_model.Accumulate(1.0, tid, &transition_accs);
	  
	  int32 k = i + shift_frame;
	  if(i + shift_frame < 0) 
	    k = 0;
	  else if (i + shift_frame >= alignment.size())
	    k = alignment.size() - 1;
	  else 
	    k = i + shift_frame;
	  
          std::vector<int32> this_gselect;
          this_gselect = (*gselect)[k];
          am_plda.ComputePerFrameVars(mat.Row(k), this_gselect, 0.0,
                                      &per_frame_vars);
          tot_like_this_file += plda_accs.Accumulate(am_plda, per_frame_vars,
                                                      pdf_id, 1.0, acc_flags);
        }

        KALDI_VLOG(2) << "Average like for this file is "
                      << (tot_like_this_file/alignment.size()) << " over "
                      << alignment.size() <<" frames.";
        tot_like += tot_like_this_file;
        tot_t += alignment.size();
        if (num_done % 50 == 0) {
          KALDI_LOG << "Processed " << num_done << " utterances; for utterance "
                    << utt << " avg. like is "
                    << (tot_like_this_file/alignment.size())
                    << " over " << alignment.size() <<" frames.";
        }
      }
    }
    KALDI_LOG << "Overall like per frame (Gaussian only) = "
              << (tot_like/tot_t) << " over " << tot_t << " frames.";

    KALDI_LOG << "Done " << num_done << " files, " << num_no_alignment
              << " with no alignments, " << num_other_error
              << " with other errors.";

    {
      Output ko(accs_wxfilename, binary);
      transition_accs.Write(ko.Stream(), binary);
      plda_accs.Write(ko.Stream(), binary);
    }
    KALDI_LOG << "Written accs.";
    return (num_done != 0 ? 0 : 1);
  } catch(const std::exception &e) {
    std::cerr << e.what();
    return -1;
  }
}


