// sgmmbin/sgmm-est.cc

// Liang Lu (2014)

// See ../../COPYING for clarification regarding multiple authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
// THIS CODE IS PROVIDED *AS IS* BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
// KIND, EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION ANY IMPLIED
// WARRANTIES OR CONDITIONS OF TITLE, FITNESS FOR A PARTICULAR PURPOSE,
// MERCHANTABLITY OR NON-INFRINGEMENT.
// See the Apache 2 License for the specific language governing permissions and
// limitations under the License.


#include "base/kaldi-common.h"
#include "util/common-utils.h"
#include "thread/kaldi-thread.h"
#include "plda/am-plda.h"
#include "hmm/transition-model.h"
#include "plda/mle-am-plda.h"


int main(int argc, char *argv[]) {
  try {
    using namespace kaldi;
    typedef kaldi::int32 int32;
    const char *usage =
        "Estimate PLDA model parameters from accumulated stats.\n"
        "Usage: plda-est [options] <model-in> <stats-in> <model-out>\n";

    bool binary_write = true;
    std::string update_flags_str = "zUGbwvt";
    std::string write_flags_str = "gsn";
    kaldi::MleTransitionUpdateConfig tcfg;
    kaldi::MleAmPldaOptions plda_opts;
    int32 split_substates = 0;
    int32 increase_phn_dim = 0;
    int32 increase_spk_dim = 0;
    BaseFloat perturb_factor = 0.01;
    BaseFloat power = 0.2;
    BaseFloat max_cond = 100;
    std::string occs_out_filename;

    ParseOptions po(usage);
    po.Register("binary", &binary_write, "Write output in binary mode");
    po.Register("split-substates", &split_substates, "Increase number of "
                "substates to this overall target.");
    po.Register("increase-phn-dim", &increase_phn_dim, "Increase phone-space "
                "dimension as far as allowed towards this target.");
    po.Register("increase-spk-dim", &increase_spk_dim, "Increase speaker-space "
                "dimension as far as allowed towards this target.");
  //  po.Register("remove-speaker-space", &remove_speaker_space, "Remove speaker-specific "
  //              "projections N");
    po.Register("power", &power, "Exponent for substate occupancies used while "
                "splitting substates.");
    po.Register("perturb-factor", &perturb_factor, "Perturbation factor for "
                "state vectors while splitting substates.");
    po.Register("max-cond-split", &max_cond, "Max condition number of smoothing "
                "matrix used in substate splitting.");
    po.Register("write-occs", &occs_out_filename, "File to write state "
                "occupancies to.");
    po.Register("update-flags", &update_flags_str, "Which PLDA parameters to "
                "update: subset of zUGbwvt.");
    po.Register("write-flags", &write_flags_str, "Which PLDA parameters to "
                "write: subset of gsn");
 //   po.Register("num-threads", &g_num_threads, "Number of threads to use in "
 //               "weight update and normalizer computation");
    tcfg.Register(&po);
    plda_opts.Register(&po);

    po.Read(argc, argv);
    if (po.NumArgs() != 3) {
      po.PrintUsage();
      exit(1);
    }
    std::string model_in_filename = po.GetArg(1),
        stats_filename = po.GetArg(2),
        model_out_filename = po.GetArg(3);

    kaldi::PldaUpdateFlagsType update_flags =
        StringToPldaUpdateFlags(update_flags_str);
    kaldi::PldaWriteFlagsType write_flags =
        StringToPldaWriteFlags(write_flags_str);
    
    AmPlda am_plda;
    TransitionModel trans_model;
    {
      bool binary;
      Input ki(model_in_filename, &binary);
      trans_model.Read(ki.Stream(), binary);
      am_plda.Read(ki.Stream(), binary);
    }

    Vector<double> transition_accs;
    MleAmPldaAccs plda_accs;
    {
      bool binary;
      Input ki(stats_filename, &binary);
      transition_accs.Read(ki.Stream(), binary);
      plda_accs.Read(ki.Stream(), binary, true);  // true == add; doesn't matter here.
    }

    if (update_flags & kPldaTransitions) {  // Update transition model.
      BaseFloat objf_impr, count;
      trans_model.MleUpdate(transition_accs, tcfg, &objf_impr, &count);
      KALDI_LOG << "Transition model update: Overall " << (objf_impr/count)
                << " log-like improvement per frame over " << (count)
                << " frames.";
    }

    { // Do the update.
      kaldi::MleAmPldaUpdater updater(plda_opts);
      updater.Update(plda_accs, &am_plda, update_flags);
    }

    if (split_substates != 0 || !occs_out_filename.empty()) {  // get state occs
      Vector<BaseFloat> state_occs;
      plda_accs.GetStateOccupancies(&state_occs);

      if (split_substates != 0) {
        am_plda.SplitSubstates(state_occs, split_substates, perturb_factor,
                               power, max_cond);
        am_plda.ComputeNormalizers();  // recompute normalizers...
      }

      if (!occs_out_filename.empty()) {
        kaldi::Output ko(occs_out_filename, false /* no binary write */);
        state_occs.Write(ko.Stream(), false  /* no binary write */);
      }
    }

    {
      Output ko(model_out_filename, binary_write);
      trans_model.Write(ko.Stream(), binary_write);
      am_plda.Write(ko.Stream(), binary_write, write_flags);
    }
    
    
    KALDI_LOG << "Written model to " << model_out_filename;
    return 0;
  } catch(const std::exception &e) {
    std::cerr << e.what();
    return -1;
  }
}


